from dataclasses import dataclass
from dataclasses_json import dataclass_json
from dataclasses_json import LetterCase
from typing import Optional


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class LinkInfo:
    path: Optional[str] = None
    object_id: Optional[int] = None
    object_type: Optional[int] = None
    share_link: Optional[str] = None
    parent_type: Optional[int] = None



from dataclasses import dataclass
from dataclasses_json import dataclass_json
from dataclasses_json import LetterCase
from typing import Optional
from datetime import datetime
from .parse import time_field
from .parse import extensions_field
from .media import Media
from .rich_format import RichFormat
from .user import User
from .circle import Circle
from .reaction import Reaction


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Blog:
    created_time: Optional[datetime] = time_field()
    edited_time: Optional[datetime] = time_field()
    updated_time: Optional[datetime] = time_field()
    blog_id: Optional[int] = None
    circle_id_list: Optional[list[int]] = None
    status: Optional[int] = None
    type: Optional[int] = None
    uid: Optional[int] = None
    content: Optional[str] = None
    rich_format: Optional[RichFormat] = None
    cover: Optional[Media] = None
    media_list: Optional[list[Media]] = None
    votes_count: Optional[int] = None
    comments_count: Optional[int] = None
    content_region: Optional[int] = None
    language: Optional[str] = None
    visibility: Optional[int] = None
    is_user_pinned: bool = False
    author: Optional[User] = False
    circle_list: Optional[list[Circle]] = None
    reaction_count_list: Optional[list[Reaction]] = None
    extensions: dict = extensions_field()



from dataclasses_json import dataclass_json
from dataclasses_json import LetterCase
from dataclasses_json import config
from dataclasses import dataclass
from dataclasses import field
from typing import Optional
from datetime import datetime
from .member_title import MemberTitle
from .media import Media
from .rich_format import RichFormat
from .sticker import Sticker
from .parse import time_field
from .parse import extensions_field


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class User:
    @dataclass_json(letter_case=LetterCase.CAMEL)
    @dataclass
    class ProfileFrame:
        resource: Optional[Media] = field(metadata=config(field_name="res"))
        created_time: Optional[datetime] = time_field()
        profile_frame_id: Optional[int] = None
        name: Optional[str] = None
        profile_frame_parent_id: Optional[int] = None
        profile_frame_parent_type: Optional[int] = None

    @dataclass_json(letter_case=LetterCase.CAMEL)
    @dataclass
    class UserMood:
        type: Optional[int] = None
        sticker_id: Optional[int] = None
        online_status: Optional[int] = None
        sticker: Optional[Sticker] = None

    @dataclass_json(letter_case=LetterCase.CAMEL)
    @dataclass
    class UserVisitorInfo:
        user_profile_visit_mode: Optional[int] = None

    created_time: Optional[datetime] = time_field()
    last_active_time: Optional[datetime] = time_field()
    uid: Optional[str] = None
    nickname: Optional[str] = None
    social_id: Optional[str] = None
    social_id_modified: Optional[str] = None
    bio: Optional[str] = None
    rich_format: Optional[RichFormat] = None
    gender: Optional[int] = None
    status: Optional[int] = None
    icon: Optional[Media] = None
    name_card_background: Optional[Media] = None
    background: Optional[Media] = None
    preview_media_list: Optional[list[Media]] = None
    profile_frame: Optional[ProfileFrame] = None
    media_list: Optional[list[Media]] = None
    chat_invitation_status: Optional[int] = None
    public_chat_invitation_status: Optional[int] = None
    private_chat_invitation_status: Optional[int] = None
    circle_invitation_status: Optional[int] = None
    online_status: Optional[int] = None
    content_region_name: Optional[str] = None
    user_mood: Optional[UserMood] = None
    content_region: Optional[int] = None
    shows_school: Optional[int] = None
    shows_location: Optional[int] = None
    name_card_enabled: Optional[int] = None
    match_enabled: Optional[int] = None
    tagline: Optional[str] = None
    zodiac_type: Optional[int] = None
    language: Optional[str] = None
    country_code: Optional[str] = None
    push_enabled: Optional[int] = None
    shows_joined_circles: Optional[int] = None
    birthday: Optional[str] = None
    comments_count: Optional[int] = None
    comment_permission_type: Optional[int] = None
    third_party_uid: Optional[str] = None
    user_visitor_info: Optional[UserVisitorInfo] = None
    wallet_activated: Optional[int] = None
    fans_count: Optional[int] = None
    friends_count: Optional[int] = None
    blogs_count: Optional[int] = None
    circle_member_title_list: list[MemberTitle] = field(default_factory=list)
    circle_joined_time: Optional[datetime] = time_field()
    circle_role: Optional[int] = None
    joined_status: Optional[int] = None
    extensions: dict = extensions_field()



from ..rich_format import RichFormat
from typing import Optional
from uuid import uuid4


class RichFormatBuilder:
    def __init__(self):
        self._fmt = RichFormat(4, [], [], [])

    def text_span(
            self,
            start: int,
            end: int,
            bold: bool = False,
            italic: bool = False,
            strikethrough: bool = False,
            underline: bool = False,
            foreground: Optional[str] = None,
            background: Optional[str] = None
    ) -> "RichFormatBuilder":
        self._fmt.text_spans.append(RichFormat.TextSpan(start, end, RichFormat.TextSpan.Data(
            bold, italic, strikethrough, underline, foreground, background
        )))
        return self

    def paragraph_span(self, start: int, end: int, style: Optional[str] = None, alignment: Optional[str] = None) -> "RichFormatBuilder":
        self._fmt.paragraph_spans.append(
            RichFormat.ParagraphSpan(start, end, RichFormat.ParagraphSpan.Data(style, alignment))
        )
        return self

    def attachment_span(self, start: int, end: int, data: RichFormat.AttachmentSpan.Data) -> "RichFormatBuilder":
        self._fmt.attachment_spans.append(
            RichFormat.AttachmentSpan(start, end, data)
        )
        return self

    def bold(self, start: int, end: int) -> "RichFormatBuilder": return self.text_span(start, end, bold=True)

    def italic(self, start: int, end: int) -> "RichFormatBuilder": return self.text_span(start, end, italic=True)

    def strikethrough(self, start: int, end: int) -> "RichFormatBuilder": return self.text_span(start, end,strikethrough=True)

    def underline(self, start: int, end: int) -> "RichFormatBuilder": return self.text_span(start, end, underline=True)

    def foreground(self, start: int, end: int, color: str) -> "RichFormatBuilder": return self.text_span(start, end, foreground=color)

    def background(self, start: int, end: int, color: str) -> "RichFormatBuilder": return self.text_span(start, end, background=color)

    def h1(self, start: int, end: int) -> "RichFormatBuilder": return self.paragraph_span(start, end, style="h1")

    def h2(self, start: int, end: int) -> "RichFormatBuilder": return self.paragraph_span(start, end, style="h2")

    def h3(self, start: int, end: int) -> "RichFormatBuilder": return self.paragraph_span(start, end, style="h3")

    def quote(self, start: int, end: int) -> "RichFormatBuilder": return self.paragraph_span(start, end, style="quote")

    def left(self, start: int, end: int) -> "RichFormatBuilder": return self.paragraph_span(start, end, alignment="left")

    def center(self, start: int, end: int) -> "RichFormatBuilder": return self.paragraph_span(start, end, alignment="center")

    def right(self, start: int, end: int) -> "RichFormatBuilder": return self.paragraph_span(start, end, alignment="right")

    def link(
        self,
        start: int,
        url: str,
        custom_title: str = "",
        title: str = "",
        media_id: int = 0
    ) -> "RichFormatBuilder":
        return self.attachment_span(start, start + 1, RichFormat.AttachmentSpan.Data(
            type="link",
            link=RichFormat.AttachmentSpan.Data.Link(url, custom_title, title, media_id, str(uuid4()))
        ))

    def mention(self, start: int, username: str, uid: int, role_id: int = 0,
                role_name_length: int = 0) -> "RichFormatBuilder":
        end = len(f"@{username}" if not username.startswith("@") else username)
        return self.attachment_span(start, end, RichFormat.AttachmentSpan.Data(
            type="mention",
            mention=RichFormat.AttachmentSpan.Data.Mention(uid, role_id, role_name_length)
        ))

    def poll(self, start: int, poll_id: int) -> "RichFormatBuilder":
        return self.attachment_span(start, start + 1, RichFormat.AttachmentSpan.Data(
            type="poll",
            poll=RichFormat.AttachmentSpan.Data.Poll(poll_id)
        ))

    def media(self, start: int, media_id: int) -> "RichFormatBuilder":
        return self.attachment_span(start, start + 1, RichFormat.AttachmentSpan.Data(
            type="media",
            media=RichFormat.AttachmentSpan.Data.Media(media_id)
        ))

    def build(self) -> RichFormat:
        return self._fmt



from dataclasses import dataclass
from dataclasses import field
from dataclasses_json import dataclass_json
from dataclasses_json import LetterCase


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class BlockUsersInfo:
    blocked_by_me_list: list[int] = field(default_factory=list)
    block_me_list: list[int] = field(default_factory=list)



from datetime import datetime
from dateutil.parser import isoparse
from typing import Union
from typing import Optional
from dataclasses import Field
from dataclasses import field
from dataclasses_json import config


def decode_time(encoded_time: Union[int, str, None]) -> Optional[datetime]:
    if isinstance(encoded_time, int):
        try: return datetime.fromtimestamp(encoded_time)
        except OSError: return datetime.fromtimestamp(encoded_time // 1000)
    elif isinstance(encoded_time, str):
        return isoparse(encoded_time)
    else: return None


def encode_time(time: Optional[datetime]) -> Optional[int]:
    if time is None:
        return None
    return int(time.timestamp() * 1000)


def decode_wallet_amount(amount: str) -> int:
    if amount == "0": return 0
    return int(amount[:len(amount) - 18])


def time_field() -> Field:
    return field(metadata=config(decoder=decode_time, encoder=encode_time), default_factory=type(None))


def extensions_field() -> Field:
    return field(default_factory=dict)


def wallet_amount_field() -> Field:
    return field(metadata=config(decoder=decode_wallet_amount), default_factory=0)



from dataclasses import dataclass
from dataclasses_json import dataclass_json
from dataclasses_json import LetterCase
from typing import Optional
from .media import Media
from .qi_vote_info import QiVoteInfo


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class QiVoteFullInfo:
    @dataclass_json(letter_case=LetterCase.CAMEL)
    @dataclass
    class VoteResources:
        enabled_icon: Optional[Media] = None
        waiting_icon: Optional[Media] = None
        disabled_icon: Optional[Media] = None
        number_background: Optional[Media] = None
        animation_icons: Optional[list[Media]] = None

    @dataclass_json(letter_case=LetterCase.CAMEL)
    @dataclass
    class VoteConfig:
        daily_quota: Optional[int] = None
        first_quota: Optional[int] = None
        waiting_time: Optional[int] = None

    vote_info: Optional[QiVoteInfo] = None
    vote_resources: Optional[VoteResources] = None
    vote_config: Optional[VoteConfig] = None



from dataclasses import dataclass
from dataclasses_json import dataclass_json
from dataclasses_json import LetterCase
from typing import Optional
from datetime import datetime
from .parse import time_field
from .media import Media


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class DefaultBackgroundMedia:
    created_time: Optional[datetime] = time_field()
    id: Optional[int] = None
    title: Optional[str] = None
    description: Optional[str] = None
    activityTitle: Optional[str] = None
    uid: Optional[int] = None
    pos: Optional[int] = None
    media: Optional[Media] = None
    from_field: Optional[int] = None



from typing import Callable
from typing import Any
from inspect import iscoroutinefunction
from asyncio import create_task
from random import randint


class SubscriptionHandler:
    def __init__(self):
        self._subscriptions = []

    def subscribe(
        self,
        handler: Callable[[Any], Any],
        content_filter: Callable[[Any], bool] = lambda _: True,
        content_transform: Callable[[Any], Any] = lambda x: x
    ) -> str:
        subscription_id = str(randint(1, 1000000))
        self._subscriptions.append({
            "id": subscription_id,
            "handler": handler,
            "filter": content_filter,
            "transform": content_transform
        })
        return subscription_id

    def unsubscribe(self, subscription_id: str) -> bool:
        try:
            self._subscriptions.remove(next(handler for handler in self._subscriptions if handler["id"] == subscription_id))
            return True
        except (StopIteration, ValueError):
            return False

    def broadcast(self, content: Any):
        for subscription in self._subscriptions:
            transformed_content = subscription["transform"](content)
            if not subscription["filter"](transformed_content):
                continue
            create_task(subscription["handler"](transformed_content)) \
                if iscoroutinefunction(subscription["handler"]) \
                else subscription["handler"](transformed_content)



from io import BytesIO
from asyncio import get_running_loop


class CopyToBufferWriter:
    def __init__(self, buffer: BytesIO):
        self.buffer = buffer

    async def write(self, data: bytes):
        await get_running_loop().run_in_executor(None, self.buffer.write, data)



from enum import Enum


class EUserTaskType(Enum):
    ACTIVATE_WALLET = 1
    BUY_MERCH = 2
    ENHANCE_PROFILE = 3
    INVITE_THREE_FRIENDS = 4
    TALK_IN_CHAT_10_MINUTES = 5
    ACTIVATE_WIDGETS = 6



from dataclasses_json import dataclass_json
from dataclasses_json import LetterCase
from dataclasses import dataclass
from typing import Optional
from datetime import datetime
from ..parse import time_field


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class MultiInvitationCodeInfo:
    uid: Optional[int] = None
    code: Optional[str] = None
    created_time: Optional[datetime] = time_field()
    last_read_time: Optional[datetime] = time_field()
    status: Optional[int] = None
    label: Optional[str] = None
    generation: Optional[int] = None
    operator_uid: Optional[int] = None
    type: Optional[int] = None
    total_used_count: Optional[int] = None
    used_count: Optional[int] = None



from dataclasses import dataclass
from dataclasses_json import dataclass_json
from dataclasses_json import LetterCase
from typing import Optional
from datetime import datetime
from .parse import time_field
from .sticker import Sticker


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Reaction:
    created_time: Optional[datetime] = time_field()
    object_id: Optional[int] = None
    sticker_id: Optional[int] = None
    object_type: Optional[int] = None
    count: Optional[int] = None
    sticker: Optional[Sticker] = None



from .api import RequestManager
from .api.headers import IHeadersProvider
from .api.headers import HeadersProvider
from .model import *
from .enum import *
from .websocket import WebsocketListener
from typing import Union
from typing import Callable
from typing import Any
from aiohttp import MultipartWriter
from typing import Optional
from aiofiles.threadpool.binary import AsyncBufferedReader
from random import randint
from sys import maxsize
from urllib.parse import urlparse
from magic import from_buffer
from mnemonic import Mnemonic
from bip32utils import BIP32Key

CircleReference = Union[Circle, str, int]


class Client(RequestManager):
    def __init__(
        self,
        commands_prefix: str = "/",
        provider: Optional[IHeadersProvider] = None,
        http_logging: bool = False,
        ws_logging: bool = False,
        *args,
        **kwargs
    ):
        super().__init__(provider or HeadersProvider(), logging=http_logging, *args, **kwargs)
        self.websocket = WebsocketListener(self, logging=ws_logging)
        self.commands_prefix = commands_prefix
        self.account = None
        self.user_profile = None

    async def _logout(self):
        self.provider.remove_sid()
        await self.websocket.disconnect()
        self.user_profile, self.account = None, None

    async def _auth(self, resp: AuthResult):
        if self.provider.get_sid(): await self._logout()
        self.provider.set_sid(resp.sid)
        self.account = resp.account
        self.user_profile = resp.user_profile
        await self.websocket.connect()

    async def _resolve_circle_reference(self, ref: CircleReference) -> int:
        if isinstance(ref, Circle):
            return ref.circle_id
        elif isinstance(ref, int) or ref.isdigit():
            return ref
        if "projz.com/s/c/" in ref:
            return (await self.get_link_info(ref)).object_id
        return (await self.get_link_info(f"https://www.projz.com/s/c/{ref}")).object_id

    async def _login(
        self,
        email: Optional[str] = None,
        phone_number: Optional[str] = None,
        password: Optional[str] = None,
        secret: Optional[str] = None
    ) -> AuthResult:
        data = {"password": password} if password is not None else {}
        if email is not None:
            data["email"] = email
            data["authType"] = EAuthType.EMAIL.value
        elif phone_number is not None:
            data["phoneNumber"] = phone_number
            data["authType"] = EAuthType.PHONE_NUMBER.value
        elif secret is not None:
            data["secret"] = secret
            data["authType"] = EAuthType.SECRET.value
            data["purpose"] = EAuthPurpose.RENEW_SID.value
        resp = AuthResult.from_dict(await self.post_json("/v1/auth/login", data))
        await self._auth(resp)
        return resp

    async def login_email(self, email: str, password: str) -> AuthResult:
        """
        Login to account using a email
        :param email: email address
        :param password: account password
        :return: model.AuthResult
        """
        return await self._login(email=email, password=password, phone_number=None)

    async def login_phone_number(self, phone_number: str, password: str) -> AuthResult:
        """
        Login to account using a phone number
        :param phone_number: phone number
        :param password: account password
        :return: model.AuthResult
        """
        return await self._login(phone_number=phone_number, password=password, email=None)

    async def login_secret(self, secret: str) -> AuthResult:
        """
        Login to account using a secret token that acts as a refresh token.
        :param secret: secret token
        :return: model.AuthResult
        """
        return await self._login(secret=secret)

    async def logout(self) -> None:
        """
        Send logout request and remove sid
        :return:
        """
        await self.post_empty("/v1/auth/logout")
        await self._logout()

    async def _request_security_validation(
        self,
        email: Optional[str],
        phone_number: Optional[str],
        purpose: Union[EAuthPurpose, int] = EAuthPurpose.LOGIN
    ) -> None:
        data = {
            "purpose": purpose.value if isinstance(purpose, EAuthPurpose) else purpose,
            "countryCode": self.country_code
        }
        if email is not None:
            data["authType"] = EAuthType.EMAIL.value
            data["email"] = email
        elif phone_number is not None:
            data["authType"] = EAuthType.PHONE_NUMBER.value
            data["phoneNumber"] = phone_number
        await self.post_json("/v1/auth/request-security-validation", data)

    async def request_email_validation(self, email: str, purpose: Union[EAuthPurpose, int] = EAuthPurpose.LOGIN) -> None:
        """
        Send verification code to email
        :param email: email address
        :param purpose: AuthPurpose.LOGIN, AuthPurpose.CHANGE_EMAIL, custom value...
        :return:
        """
        return await self._request_security_validation(email=email, purpose=purpose, phone_number=None)

    async def request_phone_validation(self, phone_number: str, purpose: Union[EAuthPurpose, int] = EAuthPurpose.LOGIN) -> None:
        """
        Send verification code to phone number
        :param phone_number: phone number
        :param purpose: AuthPurpose.LOGIN, AuthPurpose.CHANGE_PHONE, custom value...
        :return:
        """
        return await self._request_security_validation(phone_number=phone_number, purpose=purpose, email=None)

    async def change_password(self, old: str, new: str) -> None:
        """
        Change account password
        :param old: old password
        :param new: new passworod
        :return:
        """
        await self.post_json("/v1/auth/change-password", {"oldPassword": old, "newPassword": new})

    async def _check_security_validation(self, email: Optional[str], phone_number: Optional[str], code: str) -> None:
        data = {"securityCode": code}
        if email is not None:
            data["authType"] = EAuthType.EMAIL.value
            data["email"] = email
        elif phone_number is not None:
            data["authType"] = EAuthType.PHONE_NUMBER.value
            data["phoneNumber"] = phone_number
        await self.post_json("/v1/auth/check-security-validation", data)

    async def check_email_validation(self, email: str, code: str) -> None:
        """
        Check verification code sent to email
        :param email: email address
        :param code: verification code
        :return:
        """
        await self._check_security_validation(email=email, code=code, phone_number=None)

    async def check_phone_validation(self, phone_number: str, code: str) -> None:
        """
        Check verification code sent to phone number
        :param phone_number: phone number
        :param code: verification code
        :return:
        """
        await self._check_security_validation(phone_number=phone_number, code=code, email=None)

    async def _register(
        self,
        email: Optional[str],
        phone_number: Optional[str],
        password: str,
        nickname: str,
        tag_line: str,
        icon: Media,
        gender_type: EGender,
        birthday: str,
        code: str,
        name_card_background: Optional[Media] = None,
        invitation_code: Optional[str] = None,
        *,
        update_auth_credentials: bool = True
    ) -> AuthResult:
        raise NotImplementedError("[register] Not implemented -"
                                  "rawDeviceIdThree is required, "
                                  "but the correct algorithm for generating it is currently unknown.")
        # data = {
        #    "purpose": 1,
        #    "password": password,
        #    "securityCode": code,
        #    "invitationCode": invitation_code or "",
        #    "nickname": nickname,
        #    "tagLine": tag_line,
        #    "icon": icon.to_dict(),
        #    "nameCardBackground": name_card_background.to_dict() if name_card_background is not None else None,
        #    "gender": gender_type.value,
        #    "birthday": birthday,
        #    "requestToBeReactivated": False,
        #    "countryCode": self.country_code,
        #    "suggestedCountryCode": self.country_code.upper(),
        #    "ignoresDisabled": True,
        #    "rawDeviceIdThree": await self.provider.generate_device_id_three(
        #        "BU0gJ0gB5TFcCfN329Vx",
        #        "android",
        #        f"{randint(1, 12)}.{randint(1, 12)}.{randint(1, 12)}",
        #        "ASUS_Z02MQ",
        #        "default"
        #    )
        # }
        # if email is not None:
        #    data["authType"] = AuthType.EMAIL.value
        #    data["email"] = email
        # elif phone_number is not None:
        #    data["authType"] = AuthType.PHONE_NUMBER.value
        #    data["phoneNumber"] = phone_number
        # resp = await self.post_json("/v1/auth/register", data)
        # resp = AuthResult.from_dict(resp)
        # if update_auth_credentials: await self._auth(resp)
        # return AuthResult.from_dict(resp)

    async def register_email(
        self,
        email: str,
        password: str,
        nickname: str,
        tag_line: str,
        icon: Media,
        gender_type: EGender,
        birthday: str,
        code: str,
        name_card_background: Optional[Media] = None,
        invitation_code: Optional[str] = None,
        *,
        update_auth_credentials: bool = True
    ) -> AuthResult:
        """
        Create a new account with email
        :param email: email address
        :param password: account password
        :param nickname: profile nickname
        :param tag_line: profile tagline
        :param icon: icon image model.Media object
        :param gender_type: Gender.MALE | Gender.FEMALE | Gender.OTHER
        :param birthday: year-month-day
        :param code: verification code
        :param name_card_background: name card background model.Media object
        :param invitation_code: code of invitation
        :param update_auth_credentials: is it required to update client profile and sid
        :return: model.AuthResponse
        """
        return await self._register(
            email,
            None,
            password,
            nickname,
            tag_line,
            icon,
            gender_type,
            birthday,
            code,
            name_card_background,
            invitation_code,
            update_auth_credentials=update_auth_credentials
        )

    async def register_phone(
        self,
        phone_number: str,
        password: str,
        nickname: str,
        tag_line: str,
        icon: Media,
        gender_type: EGender,
        birthday: str,
        code: str,
        name_card_background: Optional[Media] = None,
        invitation_code: Optional[str] = None,
        *,
        update_auth_credentials: bool = True
    ) -> AuthResult:
        """
        Create a new account with phone number
        :param phone_number: phone number
        :param password: account password
        :param nickname: profile nickname
        :param tag_line: profile tagline
        :param icon: icon image model.Media object
        :param gender_type: Gender.MALE | Gender.FEMALE | Gender.OTHER
        :param birthday: year-month-day
        :param code: verification code
        :param name_card_background: name card background model.Media object
        :param invitation_code: code of invitation
        :param update_auth_credentials: is it required to update client profile and sid
        :return: model.AuthResponse
        """
        return await self._register(
            None,
            phone_number,
            password,
            nickname,
            tag_line,
            icon,
            gender_type,
            birthday,
            code,
            name_card_background,
            invitation_code,
            update_auth_credentials=update_auth_credentials
        )

    async def accept_chat_invitation(self, thread_id: int) -> None:
        """
        Accept invitation to the chat
        :param thread_id: id of the chat
        :return:
        """
        await self.post_empty(f"/v1/chat/threads/{thread_id}/accept-invitation")

    async def join_circle(self, reference: CircleReference) -> None:
        """
        Join to the circle
        :param reference: circle id | circle link | z id
        :return:
        """
        await self.post_empty(f"/v1/circles/{await self._resolve_circle_reference(reference)}/members")

    async def leave_circle(self, reference: CircleReference) -> None:
        """
        Leave from the chat
        :param reference: circle id | circle link | z id
        :return:
        """
        await self.delete(f"/v1/circles/{await self._resolve_circle_reference(reference)}/members")

    async def join_chat(self, thread_id: int) -> None:
        """
        Join to the chat
        :param thread_id: id of the chat
        :return:
        """
        await self.post_empty(f"/v1/chat/threads/{thread_id}/members")

    async def leave_chat(self, thread_id: int) -> None:
        """
        Leave from the chat
        :param thread_id: id of the chat
        :return:
        """
        await self.delete(f"/v1/chat/threads/{thread_id}/members")

    async def get_chat_message(self, thread_id: int, message_id: int) -> ChatMessage:
        """
        Get chat message by id
        :param thread_id: id of the chat
        :param message_id: id of the message
        :return: model.ChatMessage
        """
        return ChatMessage.from_dict(await self.get(f"/v1/chat/threads/{thread_id}/messages/{message_id}"))

    async def get_chat_messages(self, thread_id: int, size: int = 30, page_token: Optional[str] = None) -> PaginatedList[ChatMessage]:
        """
        Get list of last messages in the chat
        :param thread_id: id of the chat
        :param size: size of the list
        :param page_token: stored in PaginatedList.next_page_token
        :return: PaginatedList[model.ChatMessage]
        """
        resp = await self.get(f"/v1/chat/threads/{thread_id}/messages", {
            "size": size
        } if page_token is None else {
            "size": size,
            "pageToken": page_token
        })
        return PaginatedList(
            [ChatMessage.from_dict(message_json) for message_json in resp["list"]],
            resp.get("pagination")
        )

    async def send_message(self,
                           thread_id: int,
                           message_type: Union[EChatMessageType, int] = 1,
                           content: Optional[str] = None,
                           reply_to: Optional[int] = None,
                           message_rich_format: Optional[RichFormat] = None,
                           attached_media: Optional[Media] = None,
                           attached_audio: Optional[Media] = None,
                           poll_id: Optional[int] = None,
                           dice_id: Optional[int] = None,
                           *,
                           get_sent_message: bool = True) -> Optional[ChatMessage]:
        """
        Send message to the chat
        :param thread_id: id of the chat
        :param message_type: ChatMessageType enum field or int identifier
        :param content: text content of the message
        :param reply_to: reply message id
        :param message_rich_format: RichFormat object, can be created by RichFormatBuilder
        :param attached_media: attached to the message file model.Media object
        :param attached_audio: attached to the message audio model.Media object
        :param poll_id: attached to the message poll id
        :param dice_id: attached to the message dice id
        :param get_sent_message: If set to False, the request for information about the
        sent message will not be executed and the function will return None.
        :return:
        """
        if attached_media is not None and attached_audio is not None:
            raise ValueError("You can't send audio and media in one message")
        seq_id = randint(0, maxsize)
        data = {
            "type": message_type if isinstance(message_type, int) else message_type.value,
            "threadId": thread_id,
            "uid": self.user_profile.uid,
            "seqId": seq_id,
            "extensions": {}
        }
        if content is not None: data["content"] = content
        if reply_to is not None: data["extensions"]["replyMessageId"] = reply_to
        if message_rich_format is not None: data["richFormat"] = message_rich_format.to_dict()
        if attached_media is not None: data["media"] = attached_media.to_dict()
        if attached_audio is not None: data["media"] = attached_audio.to_dict()
        if poll_id is not None: data["extensions"]["pollId"] = poll_id
        if dice_id is not None: data["extensions"]["diceId"] = dice_id
        resp = await self.websocket.send_request(1, True, seq_id, **dict(threadId=thread_id, msg=data))
        if get_sent_message:
            return await self.get_chat_message(resp["threadId"], resp["messageId"])

    async def show_typing(self, thread_id: int) -> None:
        """
        Displays the text to the chat members that the account is "typing".
        :param thread_id: id of the chat
        :return:
        """
        await self.send_message(thread_id, EChatMessageType.TYPING, get_sent_message=False)

    async def get_block_users_info(self) -> BlockUsersInfo:
        return BlockUsersInfo.from_dict(await self.get("/v1/users/block-uids"))

    async def get_link_info(self, link: str) -> LinkInfo:
        """
        Get object info by link
        :param link: object link
        :return: model.LinkInfo
        """
        return LinkInfo.from_dict(await self.post_json("/v1/links/path", {"link": link}))

    async def parse_share_link(self, link: str) -> LinkInfo:
        """
        Web API alternative of get_link_info
        :param link: object link
        :return: model.LinkInfo
        """
        return LinkInfo.from_dict(
            await self.post_json("/v1/parse-share-link", {"link": urlparse(link).path}, web=True)
        )

    async def get_circle_info(self, circle_id: int) -> Circle:
        """
        Get circle info by id
        :param circle_id: id of the circle
        :return: model.Circle
        """
        return Circle.from_dict(await self.get(f"/v1/circles/{circle_id}"))

    async def get_chat_info(self, thread_id: int) -> Chat:
        """
        Get chat info by id
        :param thread_id: id of the chat
        :return:
        """
        return Chat.from_dict(await self.get(f"/v1/chat/threads/{thread_id}"))

    async def get_default_background_media_list(self) -> list[DefaultBackgroundMedia]:
        """
        Get default chat background list
        :return: list[model.DefaultBackgroundMedia]
        """
        return [
            DefaultBackgroundMedia.from_dict(media_json)
            for media_json in await self.get("/v1/media/default-background-media-info-list")
        ]

    async def start_chat(
        self,
        thread_type: EChatThreadType = EChatThreadType.ONE_ON_ONE,
        invite_message_content: Optional[str] = None,
        invited_user_ids: Union[str, list[str]] = None,
        circle_list: Optional[CircleReference] = None,
        title: Optional[str] = None,
        icon: Optional[Media] = None,
        background: Optional[Media] = None,
        tag_string_list: Optional[list[str]] = None,
        category_id: Optional[int] = None,
        language: Optional[str] = None
    ) -> Chat:
        """
        Create a new chat
        :param thread_type: ChatThreadType.ONE_ON_ONE | ChatThreadType.PRIVATE | ChatThreadType.PUBLIC
        :param invite_message_content: content of the auto-generated first message in the chat
        :param invited_user_ids: ids of users invited to the chat
        :param title: title of the chat
        :param icon: chat icon model.Media object
        :param background: chat background model.Media object
        :param tag_string_list: list if chat tags
        :param circle_list: circles to which the chat is attached
        :param category_id: id of the category to which the chat is attached
        :param language: language of the chat (default: en)
        :return: model.Chat
        """
        data = {
            "type": thread_type.value,
            "background": (background or (await self.get_default_background_media_list())[0].media).to_dict()
        }
        if invite_message_content is not None:
            data["inviteMessageContent"] = invite_message_content
        if invited_user_ids is not None:
            data["invitedUids"] = invited_user_ids if isinstance(invited_user_ids, list) else [invited_user_ids]
        if circle_list is not None:
            data["circleIdList"] = [await self._resolve_circle_reference(reference) for reference in circle_list]
        if title is not None: data["title"] = title
        if icon is not None: data["icon"] = icon.to_dict()
        if tag_string_list is not None: data["tagStrList"] = tag_string_list
        if category_id is not None: data["categoryId"] = category_id
        if language is not None: data["language"] = language
        return Chat.from_dict(await self.post_json("/v1/chat/threads", data))

    async def get_circle_active_members(self,
                                        reference: CircleReference,
                                        size: int = 30,
                                        page_token: Optional[str] = None) -> PaginatedList[User]:
        """
        Get active members in the circle
        :param reference: circle id | circle link | z id
        :param size: size of the list
        :param page_token: stored in PaginatedList.next_page_token
        :return: PaginatedList[model.User]
        """
        resp = await self.get(
            f"/v1/circles/{await self._resolve_circle_reference(reference)}/active-members",
            {"size": size} if page_token is None else {"pageToken": page_token, "size": size}
        )
        return PaginatedList(
            [User.from_dict(user_json) for user_json in resp["list"]],
            resp.get("pagination")
        )

    async def get_chat_with_user(self, user_id: int) -> Chat:
        """
        Try to get a one-on-one chat with another user
        :param user_id: id of the user
        :return:
        """
        return Chat.from_dict(await self.get(f"/v1/chat/one-on-one-chat/{user_id}"))

    async def get_recommended_user_namecards(self, size: int = 100, gender_type: Union[EGender, int] = 0) -> list[User]:
        """
        Get recommended user namecards
        :param size: size of the list
        :param gender_type: Gender.MALE | Gender.FEMALE | Gender.OTHER | 0
        :return: list[model.User]
        """
        resp = await self.get("/v1/users/namecards", {
            "size": size,
            "gender": gender_type if isinstance(gender_type, int) else gender_type.value,
            "withoutExtraInfo": True
        })
        return [
            User.from_dict(user_json)
            for user_json in resp["list"]
        ]

    async def get_user_info(self, user_id: int) -> User:
        """
        Get user info by id
        :param user_id: id of the user
        :return: model.User
        """
        return User.from_dict(await self.get(f"/v1/users/profile/{user_id}"))

    async def delete_chat(self, chat_id: int) -> None:
        """
        Delete a chat
        :param chat_id: id of the chat
        :return:
        """
        await self.delete(f"/v1/chat/threads/{chat_id}")

    async def get_circle_chats(
        self,
        reference: CircleReference,
        size: int = 30,
        page_token: Optional[str] = None,
        query_type: Union[EPartyQueryType, int] = EPartyQueryType.ONLINE,
        *,
        with_pin: bool = False
    ) -> PaginatedList[Chat]:
        """
        Get public chats in the circle
        :param reference: circle id | circle link | z id
        :param size: size of the list
        :param page_token: stored in PaginatedList.next_page_token
        :param with_pin: include pinned chats
        :param query_type: PartyQueryType.ALL | PartyQueryType.ONLINE
        :return: PaginatedList[model.Chat]
        """
        resp = await self.get(f"/v1/chat/threads", {
            "type": "circle",
            "objectId": await self._resolve_circle_reference(reference),
            "size": size,
            "withPin": "false",
            "partyQueryType": query_type.value if isinstance(query_type, EPartyQueryType) else query_type
        } if page_token is None else {
            "type": "circle",
            "objectId": await self._resolve_circle_reference(reference),
            "size": size,
            "withPin": with_pin,
            "partyQueryType": query_type.value if isinstance(query_type, EPartyQueryType) else query_type,
            "pageToken": page_token
        })
        return PaginatedList(
            [Chat.from_dict(user_json) for user_json in resp["list"]],
            resp.get("pagination")
        )

    async def get_joined_chats(
        self,
        start: int = 0,
        size: int = 30,
        query_type: Union[EChatQueryType, int] = EChatQueryType.PRIVATE
    ) -> list[Chat]:
        """
        Get a list of chats that an account has joined
        :param start: offset of the list
        :param size: size of the list
        :param query_type: ChatQueryType enum field or type identifier (default: private)
        :return: list[Chat]
        """
        resp = await self.get("/v1/chat/joined-threads", {
            "type": query_type.value if isinstance(query_type, EChatQueryType) else query_type,
            "start": start,
            "size": size,
        })
        return [
            Chat.from_dict(chat_json)
            for chat_json in resp["list"]
        ]

    async def get_joined_parties(self, start: int = 0, size: int = 30) -> list[Party]:
        """
        Get a list of public chats that an account has joined
        :param start: offset of the list
        :param size: size of the list
        :return: list[Party]
        """
        resp = await self.get("/v1/chat/joined-parties", {
            "type": "parties",
            "start": start,
            "size": size
        })
        return [
            Party.from_dict(party_json)
            for party_json in resp["list"]
        ]

    async def get_qi_votes_info(self, object_id: int) -> QiVoteFullInfo:
        """
        Get qi votes info
        :param object_id: id of the object
        :return: model.QiVoteFullInfo
        """
        return QiVoteFullInfo.from_dict(await self.get("/v1/qivotes", {
            "objectId": object_id,
            "timezone": self.time_zone
        }))

    async def send_qi(self, object_id: int, target_type: Union[EObjectType, int] = 1, count: int = 1) -> QiVoteInfo:
        """
        Send qi to the specified object
        :param object_id: id of the target
        :param target_type: ObjectType enum field or type identifier (default: chat)
        :param count: qi count
        :return: model.QiVoteInfo
        """
        return QiVoteInfo.from_dict(await self.post_json("/v1/qivotes", {
            "objectType": target_type if isinstance(target_type, int) else target_type.value,
            "objectId": object_id,
            "votedCount": count,
            "timezone": self.time_zone
        }))

    async def get_chat_managers(self, chat_id: int, *, only_co_hosts: bool = False) -> list[User]:
        """
        Get chat admins
        :param chat_id: id of the chat
        :param only_co_hosts: include only co-hosts
        :return: list[model.User]
        """
        resp = await self.get(
            f"/v1/chat/threads/{chat_id}/management-team",
            {"onlyCoHosts": only_co_hosts}
        )
        return [
            User.from_dict(user_json)
            for user_json in resp["list"]
        ]

    async def get_chat_members(self,
                               chat_id: int,
                               size: int = 30,
                               query_word: str = "",
                               page_token: Optional[str] = None,
                               *,
                               only_active_members: bool = False,
                               exclude_managers: bool = False) -> PaginatedList[User]:
        """
        Get list of chat members
        :param chat_id: id of the chat
        :param size: size of the list
        :param only_active_members: include only active members
        :param exclude_managers: exclude management team
        :param query_word: search query
        :param page_token: stored in PaginatedList.next_page_token
        :return: PaginatedList[model.User]
        """
        resp = await self.get(f"/v1/chat/threads/{chat_id}/members", {
            "onlyActiveMember": only_active_members,
            "size": size,
            "isExcludeManger": exclude_managers,
            "queryWord": query_word
        } if page_token is None else {
            "onlyActiveMember": only_active_members,
            "size": size,
            "isExcludeManger": exclude_managers,
            "queryWord": query_word,
            "pageToken": page_token
        })
        return PaginatedList(
            [User.from_dict(user_json) for user_json in resp["list"]],
            resp.get("pagination")
        )

    async def get_circle_members(self,
                                 reference: CircleReference,
                                 size: int = 30,
                                 query_type: ECircleMembersQueryType = ECircleMembersQueryType.NORMAL,
                                 page_token: Optional[str] = None,
                                 *,
                                 exclude_managers: bool = False) -> PaginatedList[User]:
        """
        Get list of circle members
        :param reference: circle id | circle link | z id
        :param size: size of the list
        :param query_type: CircleMembersQueryType.NORMAL | CircleMembersQueryType.BLOCKED
        :param exclude_managers: exclude management team
        :param page_token: stored in PaginatedList.next_page_token
        :return: PaginatedList[model.User]
        """
        resp = await self.get(f"/v1/circles/{await self._resolve_circle_reference(reference)}/members", {
            "type": query_type.value,
            "size": size,
            "isExcludeManger": exclude_managers
        } if page_token is None else {
            "type": query_type.value,
            "size": size,
            "isExcludeManger": exclude_managers,
            "pageToken": page_token
        })
        return PaginatedList(
            [User.from_dict(user_json) for user_json in resp["list"]],
            resp.get("pagination")
        )

    async def get_visible_member_titles(self, reference: CircleReference, size: int = 100) -> MemberTitlesInfo:
        """
        Get info about member title in specified circle
        :param reference: circle id | circle link | z id
        :param size: size of the list
        :return: model.MemberTitlesInfo
        """
        return MemberTitlesInfo.from_dict(await self.get(f"/v1/circles/{await self._resolve_circle_reference(reference)}"
                                                         f"/visible-member-titles", {"size": size}))

    async def create_poll_with_icons(
        self,
        title: str,
        poll_items: list[tuple[str, Optional[Media]]]
    ) -> Poll:
        """
        Create and register a poll with icons
        :param title: title of the poll
        :param poll_items: list of poll item contents
        :return: model.Poll
        """
        poll_items = list(map(
            lambda x: (x[0], x[1].to_dict()) if isinstance(x[1], Media) else x,
            poll_items
        ))
        return Poll.from_dict(await self.post_json("/v1/polls", {
            "title": title,
            "pollItemList": [
                {"content": item[0]} if len(item) != 2 or item[1] is None else
                {"content": item[0], "icon": item[1]}
                for item in poll_items
            ]
        }))

    async def get_categories(self,
                             target_type: Union[EObjectType, int] = 5,
                             target_region: Union[EContentRegion, int] = 1) -> list[Category]:
        """
        Get available categories
        :param target_type: ObjectType enum field (ObjectType.CIRCLE, ObjectType.CHAT, etc.) or type identifier (default: circle)
        :param target_region: categories region - ContentRegion enum field or region identifier (default: English)
        :return: list[model.Category]
        """
        resp = await self.get("/v1/categories", {
            "objectType": target_type if isinstance(target_type, int) else target_type.value,
            "contentRegion": target_region if isinstance(target_region, int) else target_region.value
        })
        return [
            Category.from_dict(category_json)
            for category_json in resp["list"]
        ]

    async def get_circles(self,
                          filter_type: Union[ECircleFilterType, str],
                          size: int = 30,
                          category_id: int = 0,
                          page_token: Optional[str] = None) -> PaginatedList[Circle]:
        """
        Get circles attached to specified category
        :param filter_type: CircleFilterType enum field or filter identifier
        :param size: size of the list
        :param category_id: id of the category, 0 for all categories
        :param page_token: stored in PaginatedList.next_page_token
        :return: PaginatedList[model.Circle]
        """
        resp = await self.get("/v1/circles", {
            "type": filter_type.value if isinstance(filter_type, ECircleFilterType) else filter_type,
            "size": size,
            "categoryId": category_id
        } if page_token is None else {
            "type": filter_type.value if isinstance(filter_type, ECircleFilterType) else filter_type,
            "size": size,
            "categoryId": category_id,
            "pageToken": page_token
        })
        return PaginatedList(
            [Circle.from_dict(circle_json) for circle_json in resp["list"]],
            resp.get("pagination")
        )

    async def visit_profile(self, user_id: int) -> None:
        """
        Visit user profile
        :param user_id: id of the user
        :return:
        """
        await self.post_empty(f"/v1/users/profile/{user_id}/visit")

    async def get_blog_info(self, blog_id: int) -> Blog:
        """
        Get blog info by blog id
        :param blog_id: id of the blog
        :return: model.Blog
        """
        return Blog.from_dict(await self.get(f"/v1/blogs/{blog_id}"))

    async def post_blog(self,
                        title: str,
                        content: str,
                        content_rich_format: Optional[RichFormat] = None,
                        media_list: Optional[list[Media]] = None,
                        background: Optional[Media] = None,
                        cover: Optional[Media] = None,
                        circle_list: Optional[list[int]] = None,
                        background_color: Optional[str] = None) -> Blog:
        """
        :param title: title of the blog
        :param content: content of the blog
        :param content_rich_format: RichFormat object, can be created by RichFormatBuilder
        :param media_list: list of model.Media objects in the blog
        :param background: blog background model.Media object
        :param cover: blog cover model.Media object
        :param circle_list: circles to which the blog is attached
        :param background_color: background color of the blog
        :return: model.Blog
        """
        circles = [await self._resolve_circle_reference(reference) for reference in circle_list] if circle_list is not None else []
        data = {
            "type": 2,
            "title": title,
            "content": content,
            "mediaList": [media_obj.to_dict() for media_obj in media_list] if media_list is not None else [],
            "richFormat": content_rich_format.to_dict() if content_rich_format is not None else RichFormatBuilder().build().to_dict(),
            "circleIdList": circles,
            "folderAffiliationList": [{"circleId": circle_id} for circle_id in circles],
            "backgroundColor": background_color or "",
            "extensions": {}
        }
        if background is not None: data["background"] = background.to_dict()
        if cover is not None: data["cover"] = cover.to_dict()
        if background_color is not None: data["extensions"]["backgroundColor"] = background_color
        return Blog.from_dict(await self.post_json("/v1/blogs", data))

    async def delete_blog(self, blog_id: int) -> None:
        """
        Delete a blog
        :param blog_id: id of the blog
        :return:
        """
        await self.delete(f"/v1/blogs/{blog_id}")

    async def get_comment_info(self, comment_id: int) -> Comment:
        """
        Get comment info by id
        :param comment_id: id of the comment
        :return: model.Comment
        """
        return Comment.from_dict(await self.get(f"/v1/comments/{comment_id}"))

    async def delete_comment(self, comment_id: int) -> Comment:
        """
        Delete a comment
        :param comment_id: id of the comment
        :return: model.Comment
        """
        await self.delete(f"/v1/comments/{comment_id}")

    async def get_comments(self,
                           parent_type: Union[EObjectType, int],
                           parent_id: int,
                           size: int = 30,
                           reply_id: int = 0,
                           page_token: Optional[str] = None,
                           *,
                           only_pinned: bool = False) -> PaginatedList[Comment]:
        """
        Get list of the comments
        :param parent_type: ObjectType enum field or type identifier
        :param parent_id: id of the parent
        :param size: size of the list
        :param reply_id: include only replies to comment with id reply_id
        :param only_pinned: include only pinned comments
        :param page_token: stored in PaginatedList.next_page_token
        :return: PaginatedList[model.Comment]
        """
        resp = await self.get("/v1/comments", {
            "parentType": parent_type if isinstance(parent_type, int) else parent_type.value,
            "parentId": parent_id,
            "size": size,
            "replyId": reply_id,
            "onlyPinned": int(only_pinned)
        } if page_token is None else {
            "parentType": parent_type if isinstance(parent_type, int) else parent_type.value,
            "parentId": parent_id,
            "size": size,
            "replyId": reply_id,
            "onlyPinned": int(only_pinned),
            "pageToken": page_token
        })
        return PaginatedList(
            [Comment.from_dict(comment_json) for comment_json in resp["list"]],
            resp.get("pagination")
        )

    async def comment(self,
                      parent_type: Union[EObjectType, int],
                      parent_id: int,
                      content: Optional[str] = None,
                      media_list: Optional[list[Media]] = None,
                      reply_to: Optional[int] = None) -> Comment:
        """
        Create a comment
        :param parent_type: ObjectType enum field or type identifier
        :param parent_id: id of the parent
        :param content: text content of the comment
        :param media_list: attachments of the comment
        :param reply_to: reply to the comment id
        :return: model.Comment
        """
        data = {
            "commentType": 1 if not media_list else 2,
            "parentType": parent_type if isinstance(parent_type, int) else parent_type.value,
            "parentId": parent_id,
            "mediaList": [media_obj.to_dict() for media_obj in media_list] if media_list is not None else []
        }
        if content is not None: data["content"] = content
        if reply_to is not None: data["replyId"] = reply_to
        return Comment.from_dict(await self.post_json("/v1/comments", data))

    async def create_poll(self, title: str, poll_items: list[str]) -> Poll:
        """
        Create and register a poll without icons
        :param title: title of the poll
        :param poll_items: list of poll item contents
        :return: model.Poll
        """
        return await self.create_poll_with_icons(title, [(item, None) for item in poll_items])

    async def get_user_tasks(self) -> list[UserTask]:
        return [
            UserTask.from_dict(task_json)
            for task_json in (await self.get("/v2/user-tasks"))["list"]
        ]

    async def accept_gift(self, gift_id: int) -> None:
        """
        Accept a gift from user or from system
        :param gift_id: id of the gift
        :return:
        """
        await self.post_empty(f"/biz/v1/gift-boxes/{gift_id}/claim")

    async def complete_user_task(self, task_type: Union[EUserTaskType, int]):
        """
        Mark user task as completed and get coins
        :param task_type: UserTaskType enum field or custom int value
        :return:
        """
        await self.get_user_tasks()
        resp = await self.post_json("/v1/user-tasks/claim-reward", {
            "type": task_type.value if isinstance(task_type, EUserTaskType) else task_type
        })
        await self.accept_gift(resp["giftBox"]["boxId"])

    async def send_gift(
        self,
        user_id: int,
        count: int,
        payment_password: str,
        title: str,
        sending_currency_type: ECurrencyType
    ) -> GiftBox:
        """
        Send gift to user
        :param user_id: target user id
        :param count: count of coins/diamonds/merch
        :param payment_password: password needed to perform operation
        :param title: attached message
        :param sending_currency_type: CurrencyType enum value
        :return: model.GiftBox
        """
        resp = await self.post_json("/biz/v1/gift-boxes", {
            "toObjectId": user_id,
            "toObjectType": EObjectType.USER.value,
            "amount": str(count) + ("0" * 18),
            "paymentPassword": payment_password,
            "currencyType": sending_currency_type.value,
            "title": title
        })
        return GiftBox.from_dict(resp)

    async def get_transfer_orders(self, size: int = 30, page_token: Optional[str] = None) -> PaginatedList[TransferOrder]:
        """
        Get a list of incoming currency transfers
        :param size: size of the list
        :param page_token: stored in PaginatedList.next_page_token
        :return: PaginatedList[model.TransferOrder]
        """
        resp = await self.get("/biz/v2/transfer-orders", {
            "size": size
        } if page_token is None else {
            "size": size,
            "pageToken": page_token
        })
        return PaginatedList(
            [TransferOrder.from_dict(order_json) for order_json in resp["list"]],
            resp.get("pagination")
        )

    async def activate_wallet(self, payment_password: str, security_code: str) -> str:
        """
        Activate a wallet
        :param payment_password: password for performing payment operations
        :param security_code: validation code
        :return: recovery phrase
        """
        await self.get_wallet_info()
        
        mnemonic = Mnemonic(language="english")
        recovery_phrase = mnemonic.generate(strength=128)
        await self.post_json("/biz/v1/wallet/0/activate", {
            "paymentPassword": payment_password,
            "securityCode": security_code,
            "identity": self.account.email or self.account.phone_number,
            "authType": EAuthType.EMAIL.value if self.account.email is not None else EAuthType.PHONE_NUMBER.value,
            "recoveryPhrasePublicKey": BIP32Key.fromEntropy(mnemonic.to_seed(recovery_phrase)).PublicKey().hex()
        })
        return recovery_phrase

    async def get_wallet_info(self) -> Optional[Wallet]:
        """
        Get info about user's wallet
        :return: model.Wallet
        """
        resp = await self.get("/biz/v1/wallet")
        if len(resp) == 0: return None
        return Wallet.from_dict(resp)

    async def get_dices(self) -> list[Dice]:
        """
        Get list of currently available dices
        :return list[model.Dice]
        """
        return [
            Dice.from_dict(dice_json)
            for dice_json in (await self.get("/v1/dices"))["diceList"]
        ]

    async def get_invitation_code(self) -> MultiInvitationCodeInfo:
        """
        Get info about new user invitation code
        :return: model.MultiInvitationCodeInfo
        """
        return MultiInvitationCodeInfo.from_dict(await self.get("/v1/users/multi-invitation-code"))

    async def follow(self, user_id: int) -> None:
        """
        Follow user profile
        :param user_id: target user id
        :return:
        """
        await self.post_empty(f"/v1/users/membership/{user_id}")

    async def unfollow(self, user_id: int) -> None:
        """
        Unfollow user profile
        :param user_id: target user id
        :return:
        """
        await self.delete(f"/v1/users/membership/{user_id}")

    async def get_blocked_items(
        self,
        reference: CircleReference,
        items_type: Union[EObjectType, int],
        size: int = 30,
        page_token: Optional[str] = None
    ) -> PaginatedList[BlockedItemWrapper]:
        """
        Get blocked items in the circle
        :param reference: circle id | circle link | z id
        :param items_type: ObjectType enum field or type identifier
        :param size: size of the list
        :param page_token: stored in PaginatedList.next_page_token
        :return: PaginatedList[BlockedItemWrapper]
        """
        resp = await self.get(f"/v1/circles/{await self._resolve_circle_reference(reference)}/blocked-items", {
            "objectType": items_type.value if isinstance(items_type, EObjectType) else items_type,
            "size": size
        } if page_token is None else {
            "objectType": items_type.value if isinstance(items_type, EObjectType) else items_type,
            "size": size,
            "page_token": page_token
        })
        return PaginatedList(
            [BlockedItemWrapper.from_dict(item_json) for item_json in resp["list"]],
            resp.get("pagination")
        )

    async def get_blocked_blogs(
        self,
        reference: CircleReference,
        size: int = 30,
        page_token: Optional[str] = None
    ) -> PaginatedList[Blog]:
        """
        Get blogs blocked in the selected circle
        :param reference: circle id | circle link | z id
        :param size: size of the list
        :param page_token: stored in PaginatedList.next_page_token
        :return: PaginatedList[Blog]
        """
        wrappers = await self.get_blocked_items(reference, EObjectType.BLOG, size, page_token)
        return PaginatedList(
            list(map(
                lambda x: x.blog,
                filter(lambda x: x.blog is not None, wrappers)
            )),
            next_page_token=wrappers.next_page_token
        )

    async def get_blocked_chats(
        self,
        reference: CircleReference,
        size: int = 30,
        page_token: Optional[str] = None
    ) -> PaginatedList[Chat]:
        """
        Get chats blocked in the selected circle
        :param reference: circle id | circle link | z id
        :param size: size of the list
        :param page_token: stored in PaginatedList.next_page_token
        :return: PaginatedList[Chat]
        """
        wrappers = await self.get_blocked_items(reference, EObjectType.CHAT, size, page_token)
        return PaginatedList(
            list(map(
                lambda x: x.chat,
                filter(lambda x: x.chat is not None, wrappers)
            )),
            next_page_token=wrappers.next_page_token
        )

    async def block_item(self, reference: CircleReference, item_id: int, item_type: Union[EObjectType, int]) -> None:
        """
        Block item in the selected circle
        :param reference: circle id | circle link | z id
        :param item_id: id of the item
        :param item_type: ObjectType enum field or type identifier
        :return:
        """
        await self.post_json(f"/v1/circles/{await self._resolve_circle_reference(reference)}/blocked-items", {
            "objectId": item_id,
            "objectType": item_type.value if isinstance(item_type, EObjectType) else item_type
        })

    async def unblock_item(self, reference: CircleReference, item_id: int) -> None:
        """
        Unblock item in the selected circle
        :param reference: circle id | circle link | z id
        :param item_id: id of the item
        :return:
        """
        await self.delete(f"/v1/circles/{await self._resolve_circle_reference(reference)}/blocked-items/{item_id}")

    async def change_chat_online_status(self, chat_id: int, *, is_online: bool) -> None:
        """
        Enable or disable the chat
        :param chat_id: id of the chat
        :param is_online: True | False
        :return:
        """
        await self.post_json(f"/v1/chat/threads/{chat_id}/party-online-status", {
            "partyOnlineStatus": 1 if is_online else 2
        })

    async def remove_circle_member(
        self,
        reference: CircleReference,
        member_id: int,
        *,
        block_member: bool,
        remove_content: bool
    ) -> None:
        """
        Kick or ban circle member
        :param reference: circle id | circle link | z id
        :param member_id: id of the member
        :param block_member: is it required to ban member
        :param remove_content: is it required to remove all content posted by member
        :return:
        """
        await self.post_json(f"/v1/circles/{await self._resolve_circle_reference(reference)}/members/{member_id}", {
            "type": "block" if block_member else "remove",
            "removeContent": remove_content
        })

    async def kick_circle_member(
        self,
        reference: CircleReference,
        member_id: int,
        *,
        remove_content: bool = False
    ) -> None:
        """
        Kick circle member
        :param reference: circle id | circle link | z id
        :param member_id: id of the member
        :param remove_content: is it required to remove all content posted by member
        :return:
        """
        await self.remove_circle_member(reference, member_id, block_member=False, remove_content=remove_content)

    async def ban_circle_member(
        self,
        reference: CircleReference,
        member_id: int,
        *,
        remove_content: bool = False
    ) -> None:
        """
        Ban circle member
        :param reference: circle id | circle link | z id
        :param member_id: id of the member
        :param remove_content: is it required to remove all content posted by member
        :return:
        """
        await self.remove_circle_member(reference, member_id, block_member=True, remove_content=remove_content)

    async def unban_circle_member(self, reference: CircleReference, member_id: int) -> None:
        """
        Unban circle member
        :param reference: circle id | circle link | z id
        :param member_id: id of the member
        :return:
        """
        await self.post_json(f"/v1/circles/{await self._resolve_circle_reference(reference)}/members/{member_id}", {
            "type": "unblock"
        })

    async def remove_chat_member(
        self,
        chat_id: int,
        member_id: int,
        *,
        block_member: bool,
        remove_content: bool
    ) -> None:
        """
        Kick or ban chat member
        :param chat_id: id of the chat
        :param member_id: id of the member
        :param block_member: is it required to ban member
        :param remove_content: is it required to remove all content sent by member
        :return:
        """
        await self.delete(f"/v1/chat/threads/{chat_id}/members/{member_id}", {
            "block": block_member,
            "removeContent": remove_content
        })

    async def kick_chat_member(self, chat_id: int, member_id: int, *, remove_content: bool = False) -> None:
        """
        Kick chat member
        :param chat_id: id of the chat
        :param member_id: id of the member
        :param remove_content: is it required to remove all content posted by member
        :return:
        """
        await self.remove_chat_member(chat_id, member_id, block_member=False, remove_content=remove_content)

    async def ban_chat_member(self, chat_id: int, member_id: int, *, remove_content: bool = False) -> None:
        """
        Ban chat member
        :param chat_id: id of the chat
        :param member_id: id of the member
        :param remove_content: is it required to remove all content posted by member
        :return:
        """
        await self.remove_chat_member(chat_id, member_id, block_member=True, remove_content=remove_content)

    async def invite_to_chat(self, chat_id: int, invited_users: Union[str, list[str]]) -> None:
        """
        Invite a new user to the chat
        :param chat_id: id of the chat
        :param invited_users: id(s) of the user(s)
        :return:
        """
        await self.post_json(f"/v1/chat/threads/{chat_id}/members-invite", {
            "invitedUids": [invited_users] if isinstance(invited_users, str) else invited_users
        })

    async def delete_chat_message(self, chat_id: int, message_id: int) -> None:
        """
        Delete a message from the chat
        :param chat_id: id of the chat
        :param message_id: id of the message
        :return:
        """
        await self.delete(f"/v1/chat/threads/{chat_id}/messages/{message_id}")

    async def upload_file(self,
                          file: Union[bytes, AsyncBufferedReader],
                          target: Union[EUploadTarget, int],
                          duration: int = 0,
                          *,
                          raw_output: bool = False) -> Union[dict, Media]:
        """
        Upload a file to the Project Z server
        :param file: file object returned by aiofiles.open or raw in-memory byte buffer
        :param target: UploadTarget enum field or int identifier
        :param duration: audio duration in milliseconds or 0
        :param raw_output: is it required to return dictionary object instead of model.Media
        :return: dict | model.Media
        """
        file_content = await file.read() if isinstance(file, AsyncBufferedReader) else file
        target = target if isinstance(target, int) else target.value
        writer = MultipartWriter()
        part = writer.append(file_content, {"Content-Type": from_buffer(file_content, mime=True)})
        part.set_content_disposition(
            "form-data",
            name="media",
            filename=file.name if isinstance(file, AsyncBufferedReader) else "file"
        )
        resp = await self.post(f"/v1/media/upload?target={target}&duration={duration}",
                               writer,
                               content_type=f"multipart/form-data; boundary={writer.boundary}")
        return resp if raw_output else Media.from_dict(resp)

    def on_message(self, text: Optional[str] = None):
        """
        A high-level decorator for registering message handlers with a specific text.
        :param text: text of the messages
        :return:
        """
        def decorator(handler: Callable[[ChatMessage], Any]):
            self.register_chat_message_handler(handler, lambda x: (x.content == text) if text is not None else True)

        return decorator

    def on_command(self, text: str, prefix: Optional[str] = None):
        """
        A high-level decorator for registering message handlers with a commands with a specific text.
        The command differs from the message in that it has a prefix.
        :param text: text of the messages
        :param prefix: The prefix of the command. By default, the prefix passed to the Client arguments is used.
        :return:
        """
        return self.on_message(f"{prefix or self.commands_prefix}{text}")

    def register_chat_message_handler(
        self,
        handler: Callable[[ChatMessage], Any],
        content_filter: Callable[[ChatMessage], bool] = lambda _: True,
        content_transform: Callable[[ChatMessage], Any] = lambda x: x
    ) -> None:
        """
        Register handler to receive new message events from the server
        :param handler: function(model.ChatMessage) -> Any
        :param content_filter: function(model.ChatMessage) -> bool
        :param content_transform: function(model.ChatMessage) -> Any
        :return:
        """
        self.websocket.subscribe(handler, lambda x: isinstance(x, ChatMessage) and content_filter(x), content_transform)



from .response import AuthResult
from .response import MemberTitlesInfo
from .response import MultiInvitationCodeInfo
from .response import BlockUsersInfo
from .util import PaginatedList
from .util import RichFormatBuilder
from .util import RichFormatTextBuilder
from .media import Media
from .account import Account
from .user import User
from .link_info import LinkInfo
from .sticker import Sticker
from .category import Category
from .circle import Circle
from .reaction import Reaction
from .blog import Blog
from .chat import Chat
from .default_background_media import DefaultBackgroundMedia
from .chat_message import ChatMessage
from .poll import Poll
from .dice import Dice
from .rich_format import RichFormat
from .qi_vote_info import QiVoteInfo
from .qi_vote_full_info import QiVoteFullInfo
from .member_title import MemberTitle
from .comment import Comment
from .currency import Currency
from .user_task import UserTask
from .wallet import Wallet
from .gift_box import GiftBox
from .transfer_order import TransferOrder
from .party import Party
from .blocked_item_wrapper import BlockedItemWrapper



from dataclasses import dataclass
from dataclasses_json import dataclass_json
from dataclasses_json import LetterCase
from .media import Media


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Sticker:
    sticker_id: int
    name: str
    media: Media



from dataclasses import dataclass
from dataclasses_json import dataclass_json
from dataclasses_json import LetterCase
from typing import Optional
from datetime import datetime
from .parse import time_field
from .parse import extensions_field
from .user import User
from .media import Media
from .rich_format import RichFormat
from .poll import Poll
from .dice import Dice


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ChatMessage:
    created_time: Optional[datetime] = time_field()
    thread_id: Optional[int] = None
    message_id: Optional[int] = None
    uid: Optional[int] = None
    type: Optional[int] = None
    seq_id: Optional[int] = None
    content: Optional[str] = None
    author: Optional[User] = None
    media: Optional[Media] = None
    poll: Optional[Poll] = None
    dice: Optional[Dice] = None
    rich_format: Optional[RichFormat] = None
    extensions: dict = extensions_field()



from dataclasses import dataclass
from dataclasses_json import dataclass_json
from dataclasses_json import LetterCase
from typing import Optional
from datetime import datetime
from .parse import time_field
from .user import User
from .media import Media


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Comment:
    created_time: Optional[datetime] = time_field()
    comment_id: Optional[int] = None
    parent_type: Optional[int] = None
    parent_id: Optional[int] = None
    uid: Optional[int] = None
    status: Optional[int] = None
    content: Optional[str] = None
    comment_type: Optional[int] = None
    media_list: Optional[list[Media]] = None
    votes_count: Optional[int] = None
    author: Optional[User] = None



from .rich_format_builder import RichFormatBuilder
from ..rich_format import RichFormat
from typing import Optional


class RichFormatTextBuilder:
    def __init__(self):
        self._builder = RichFormatBuilder()
        self._previous_offset = 0
        self._current_offset = 0
        self._text = ""

    def _update_state(self, text: Optional[str] = None) -> "RichFormatTextBuilder":
        if text is not None:
            self._text += text
        self._previous_offset = self._current_offset
        self._current_offset = len(self._text)
        return self

    def _update_increment(self) -> "RichFormatTextBuilder":
        self._current_offset += 1
        return self

    def bold(self, text: Optional[str] = None) -> "RichFormatTextBuilder":
        self._builder.bold(self._current_offset, self._current_offset + len(text or self._text))
        return self._update_state(text)

    def italic(self, text: Optional[str] = None) -> "RichFormatTextBuilder":
        self._builder.italic(self._current_offset, self._current_offset + len(text or self._text))
        return self._update_state(text)

    def strikethrough(self, text: Optional[str] = None) -> "RichFormatTextBuilder":
        self._builder.strikethrough(self._current_offset, self._current_offset + len(text or self._text))
        return self._update_state(text)

    def underline(self, text: Optional[str] = None) -> "RichFormatTextBuilder":
        self._builder.underline(self._current_offset, self._current_offset + len(text or self._text))
        return self._update_state(text)

    def foreground(self, color: str, text: Optional[str] = None) -> "RichFormatTextBuilder":
        self._builder.foreground(self._current_offset, self._current_offset + len(text or self._text), color)
        return self._update_state(text)

    def background(self, color: str, text: Optional[str] = None) -> "RichFormatTextBuilder":
        self._builder.background(self._current_offset, self._current_offset + len(text or self._text), color)
        return self._update_state(text)

    def h1(self, text: Optional[str] = None) -> "RichFormatTextBuilder":
        self._builder.h1(self._current_offset, self._current_offset + len(text or self._text))
        return self._update_state(text)

    def h2(self, text: Optional[str] = None) -> "RichFormatTextBuilder":
        self._builder.h2(self._current_offset, self._current_offset + len(text or self._text))
        return self._update_state(text)

    def h3(self, text: Optional[str] = None) -> "RichFormatTextBuilder":
        self._builder.h3(self._current_offset, self._current_offset + len(text or self._text))
        return self._update_state(text)

    def quote(self, text: Optional[str] = None) -> "RichFormatTextBuilder":
        self._builder.quote(self._current_offset, self._current_offset + len(text or self._text))
        return self._update_state(text)

    def left(self, text: Optional[str] = None) -> "RichFormatTextBuilder":
        self._builder.left(self._current_offset, self._current_offset + len(text or self._text))
        return self._update_state(text)

    def center(self, text: Optional[str] = None) -> "RichFormatTextBuilder":
        self._builder.center(self._current_offset, self._current_offset + len(text or self._text))
        return self._update_state(text)

    def right(self, text: Optional[str] = None) -> "RichFormatTextBuilder":
        self._builder.right(self._current_offset, self._current_offset + len(text or self._text))
        return self._update_state(text)

    def link(self, url: str, custom_title: str = "", title: str = "", media_id: int = 0) -> "RichFormatTextBuilder":
        self._builder.link(self._current_offset, url, custom_title, title, media_id)
        return self._update_increment()

    def mention(self, username: str, uid: int, role_id: int = 0, role_name_length: int = 0) -> "RichFormatTextBuilder":
        self._builder.mention(self._current_offset, username, uid, role_id, role_name_length)
        return self._update_increment()

    def poll(self, poll_id: int) -> "RichFormatTextBuilder":
        self._builder.poll(self._current_offset, poll_id)
        return self._update_increment()

    def media(self, media_id: int) -> "RichFormatTextBuilder":
        self._builder.media(self._current_offset, media_id)
        return self._update_increment()

    def newline(self) -> "RichFormatTextBuilder":
        self._update_state("\n")
        return self

    def back(self) -> "RichFormatTextBuilder":
        self._current_offset = self._previous_offset
        return self

    def build(self) -> tuple[RichFormat, str]:
        return self._builder.build(), self._text



from dataclasses import dataclass
from dataclasses_json import dataclass_json
from dataclasses_json import LetterCase
from typing import Optional
from datetime import datetime
from .parse import time_field
from .parse import extensions_field
from .media import Media
from .user import User
from .category import Category


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Circle:
    @dataclass_json(letter_case=LetterCase.CAMEL)
    @dataclass
    class CircleBackground:
        background_image: Optional[Media] = None
    created_time: Optional[datetime] = time_field()
    updated_time: Optional[datetime] = time_field()
    circle_id: Optional[int] = None
    category_id: Optional[int] = None
    concept_id: Optional[int] = None
    category: Optional[Category] = None
    social_id: Optional[str] = None
    social_id_modified: Optional[int] = None
    status: Optional[int] = None
    verified_status: Optional[int] = None
    name: Optional[str] = None
    tagline: Optional[str] = None
    language: Optional[str] = None
    content_region: Optional[int] = None
    members_count: Optional[int] = None
    daily_active_user: Optional[int] = None
    daily_new_post_count: Optional[int] = None
    privacy: Optional[int] = None
    join_permission: Optional[int] = None
    visibility: Optional[int] = None
    discoverability: Optional[int] = None
    icon: Optional[Media] = None
    cover: Optional[Media] = None
    circle_icon: Optional[Media] = None
    circle_background: Optional[CircleBackground] = None
    theme_color: Optional[str] = None
    joined_status: Optional[int] = None
    uid: Optional[int] = None
    author: Optional[User] = None
    background: Optional[Media] = None
    admin_id_list: Optional[list[int]] = None
    co_admin_id_list: Optional[list[int]] = None
    root_folder_id: Optional[int] = None
    description: Optional[str] = None
    # rich format
    # home tab background
    # home layout
    # menu
    # current member info
    extensions: dict = extensions_field()



from dataclasses import dataclass
from dataclasses_json import dataclass_json
from dataclasses_json import LetterCase
from typing import Optional
from datetime import datetime
from .parse import time_field
from .media import Media
from .rich_format import RichFormat
from .category import Category
from .user import User
from .chat_message import ChatMessage
from .blog import Blog
from .circle import Circle


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Chat:
    @dataclass_json(letter_case=LetterCase.CAMEL)
    @dataclass
    class EventTag:
        endTime: Optional[datetime] = time_field()
        tagId: Optional[int] = None
        tagName: Optional[str] = None
        media: Optional[Media] = None
        status: Optional[int] = None
    created_time: Optional[datetime] = time_field()
    updated_time: Optional[datetime] = time_field()
    thread_id: Optional[int] = None
    type: Optional[int] = None
    host_uid: Optional[int] = None
    co_host_uids: Optional[list[int]] = None
    title: Optional[str] = None
    icon: Optional[Media] = None
    background: Optional[Media] = None
    content: Optional[str] = None
    rich_format: Optional[RichFormat] = None
    latest_message_id: Optional[int] = None
    members_count: Optional[int] = None
    all_members_count: Optional[int] = None
    content_region: Optional[int] = None
    category_id: Optional[int] = None
    concept_id: Optional[int] = None
    category: Optional[Category] = None
    welcome_message: Optional[str] = None
    members_summary: Optional[list[User]] = None
    language: Optional[str] = None
    visibility: Optional[int] = None
    roles_count: Optional[int] = None
    # current member info
    latest_message: Optional[ChatMessage] = None
    host: Optional[User] = None
    revolving_lantern_blog: Optional[Blog] = None
    circle_id_list: Optional[list[int]] = None
    circle_list: Optional[list[Circle]] = None
    qi_voted_count: Optional[int] = None
    event_tag: Optional[EventTag] = None
    using_role_count: Optional[int] = None
    talking_member_count: Optional[int] = None
    roleplayer_count: Optional[int] = None
    # mentionedBy
    plt: Optional[int] = None



from dataclasses import dataclass
from dataclasses_json import dataclass_json
from dataclasses_json import LetterCase
from typing import Optional


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class MemberTitle:
    title_id: Optional[int] = None
    circle_id: Optional[int] = None
    title: Optional[str] = None
    description: Optional[str] = None
    color: Optional[str] = None
    type: Optional[int] = None
    visible: bool = False
    position: Optional[int] = None
    status: Optional[int] = None



from dataclasses import dataclass
from dataclasses_json import dataclass_json
from dataclasses_json import LetterCase
from typing import Optional


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class RichFormat:
    @dataclass_json(letter_case=LetterCase.CAMEL)
    @dataclass
    class TextSpan:
        @dataclass_json(letter_case=LetterCase.CAMEL)
        @dataclass
        class Data:
            bold: Optional[bool] = None
            italic: Optional[bool] = None
            strikethrough: Optional[bool] = None
            underline: Optional[bool] = None
            foreground_color: Optional[str] = None
            background_color: Optional[str] = None
        start: Optional[int] = None
        end: Optional[int] = None
        data: Optional[Data] = None

    @dataclass_json(letter_case=LetterCase.CAMEL)
    @dataclass
    class ParagraphSpan:
        @dataclass_json(letter_case=LetterCase.CAMEL)
        @dataclass
        class Data:
            style: Optional[str] = None
            alignment: Optional[str] = None
        start: Optional[int] = None
        end: Optional[int] = None
        data: Optional[Data] = None

    @dataclass_json(letter_case=LetterCase.CAMEL)
    @dataclass
    class AttachmentSpan:
        @dataclass_json(letter_case=LetterCase.CAMEL)
        @dataclass
        class Data:
            @dataclass_json(letter_case=LetterCase.CAMEL)
            @dataclass
            class Link:
                url: Optional[str] = None
                custom_title: Optional[str] = None
                title: Optional[str] = None
                media_ref_id: Optional[int] = None
                editing_media_id: Optional[str] = None

            @dataclass_json(letter_case=LetterCase.CAMEL)
            @dataclass
            class Mention:
                uid: Optional[int] = None
                role_id: Optional[int] = None
                role_name_length: Optional[int] = None

            @dataclass_json(letter_case=LetterCase.CAMEL)
            @dataclass
            class Poll:
                poll_ref_id: Optional[int] = None

            @dataclass_json(letter_case=LetterCase.CAMEL)
            @dataclass
            class Media:
                media_ref_id: Optional[int] = None
            type: Optional[str] = None
            link: Optional[Link] = None
            mention: Optional[Mention] = None
            media: Optional[Media] = None
            poll: Optional[Poll] = None
            # other
        start: Optional[int] = None
        end: Optional[int] = None
        data: Optional[Data] = None
    version: Optional[int] = None
    text_spans: Optional[list[TextSpan]] = None
    paragraph_spans: Optional[list[ParagraphSpan]] = None
    attachment_spans: Optional[list[AttachmentSpan]] = None



from dataclasses import dataclass
from dataclasses_json import dataclass_json
from dataclasses_json import LetterCase
from typing import Optional
from datetime import datetime
from .parse import time_field
from .parse import extensions_field
from .sticker import Sticker


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Category:
    @dataclass_json(letter_case=LetterCase.CAMEL)
    @dataclass
    class TagInfo:
        @dataclass_json(letter_case=LetterCase.CAMEL)
        @dataclass
        class Style:
            background_color: Optional[str] = None
            text_color: Optional[str] = None
            border_color: Optional[str] = None
            solid_color: Optional[str] = None
        created_time: Optional[datetime] = time_field()
        tag_id: Optional[int] = None
        tag_name: Optional[str] = None
        title: Optional[str] = None
        lower_case_name: Optional[str] = None
        language_code: Optional[str] = None
        status: Optional[int] = None
        official_verified: Optional[int] = None
        tag_type: Optional[int] = None
        style: Optional[Style] = None
    created_time: Optional[datetime] = time_field()
    category_id: Optional[int] = None
    sticker_id: Optional[int] = None
    name: Optional[str] = None
    title: Optional[str] = None
    lower_case_name: Optional[str] = None
    content_region: Optional[int] = None
    status: Optional[int] = None
    object_type: Optional[int] = None
    score: Optional[int] = None
    sticker: Optional[Sticker] = None
    concept_id: Optional[int] = None
    concept_name: Optional[str] = None
    category_concept: Optional[int] = None
    tag_info: Optional[TagInfo] = None
    extensions: dict = extensions_field()



from dataclasses_json import dataclass_json
from dataclasses_json import LetterCase
from dataclasses import dataclass
from typing import Optional
from datetime import datetime
from .parse import time_field


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Account:
    uid: Optional[int] = None
    status: Optional[int] = None
    email: Optional[str] = None
    phone_number: Optional[str] = None
    device_id: Optional[str] = None
    has_profile: Optional[int] = None
    has_password: Optional[int] = None
    current_device_id: Optional[str] = None
    current_device_id2: Optional[str] = None
    current_device_id3: Optional[str] = None
    registered_device_id: Optional[str] = None
    registered_device_id2: Optional[str] = None
    registered_device_id3: Optional[str] = None
    registered_ipv4: Optional[str] = None
    last_login_ipv4: Optional[str] = None
    created_time: Optional[datetime] = time_field()



from dataclasses_json import dataclass_json
from dataclasses_json import LetterCase
from dataclasses_json import config
from dataclasses import dataclass
from dataclasses import field
from ..account import Account
from ..user import User


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class AuthResult:
    secret: str
    account: Account
    user_profile: User
    sid: str = field(metadata=config(field_name="sId"))



from dataclasses import dataclass
from dataclasses import field
from dataclasses_json import dataclass_json
from dataclasses_json import LetterCase
from datetime import datetime
from typing import Optional
from .currency import Currency
from .parse import extensions_field
from .parse import time_field
from .parse import wallet_amount_field


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Wallet:
    @dataclass_json(letter_case=LetterCase.CAMEL)
    @dataclass
    class WalletAccount:
        @dataclass_json(letter_case=LetterCase.CAMEL)
        @dataclass
        class WalletCurrencyListItem:
            account_id: Optional[int] = None
            currency_type: Optional[int] = None
            balance: Optional[Currency] = None
            available_currency: Optional[Currency] = None
            balance_value: Optional[int] = wallet_amount_field()
            created_time: Optional[datetime] = time_field()
        account_id: Optional[int] = None
        account_type: Optional[int] = None
        status: Optional[int] = None
        owner_id: Optional[int] = None
        owner_type: Optional[int] = None
        created_time: Optional[datetime] = time_field()
        currency_list: list[WalletCurrencyListItem] = field(default_factory=list)

    uid: Optional[int] = None
    activate_status: Optional[int] = None
    status: Optional[int] = None
    payment_password_version: Optional[int] = None
    regular_account: Optional[WalletAccount] = None
    last_read_transfer_order_id: Optional[int] = None
    created_time: Optional[datetime] = time_field()
    wallet_extensions: dict = extensions_field()



from dataclasses import dataclass
from dataclasses_json import dataclass_json
from dataclasses_json import LetterCase
from typing import Optional
from datetime import datetime
from .parse import time_field


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class QiVoteInfo:
    created_time: Optional[datetime] = time_field()
    last_vote_time: Optional[datetime] = time_field()
    object_type: Optional[int] = None
    uid: Optional[int] = None
    voted_date: Optional[int] = None
    voted_count: Optional[int] = None
    total_voted_count: Optional[int] = None



from dataclasses import dataclass
from dataclasses import field
from dataclasses_json import dataclass_json
from dataclasses_json import config
from dataclasses_json import LetterCase
from ..user import User
from ..member_title import MemberTitle


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class MemberTitlesInfo:
    titles: list[MemberTitle] = field(metadata=config(field_name="list"), default_factory=list)
    member_list_map: dict[str, list[User]] = field(default_factory=dict)



from dataclasses import dataclass
from dataclasses_json import dataclass_json
from dataclasses_json import LetterCase
from typing import Optional
from datetime import datetime
from .parse import time_field
from .parse import extensions_field


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Poll:
    @dataclass_json(letter_case=LetterCase.CAMEL)
    @dataclass
    class PollItem:
        created_time: Optional[datetime] = time_field()
        poll_id: Optional[int] = None
        poll_item_id: Optional[int] = None
        content: Optional[str] = None
    created_time: Optional[datetime] = time_field()
    poll_id: Optional[int] = None
    uid: Optional[int] = None
    title: Optional[str] = None
    poll_item_list: Optional[list[PollItem]] = None
    extensions: dict = extensions_field()



from abc import ABC


class IHeadersProvider(ABC):
    def get_persistent_headers(self) -> dict: ...

    def get_request_info_headers(self,
                                 device_id: str,
                                 nonce: str,
                                 language: str,
                                 country_code: str,
                                 time_zone: int) -> dict: ...

    def get_signable_header_keys(self) -> list[str]: ...

    def get_sid(self) -> str: ...

    def set_sid(self, sid: str) -> None: ...

    def remove_sid(self) -> None: ...

    async def generate_request_signature(self, path: dict, headers: dict, body: bytes) -> str: ...

    async def generate_device_id(self, installation_id: str) -> str: ...

    async def generate_device_id_three(
        self,
        organization: str,
        platform: str,
        version: str,
        model: str,
        app_id: str
    ) -> str: ...



from enum import Enum


class EUploadTarget(Enum):
    USER_ICON = 1
    BACKGROUND = 2
    FOREGROUND = 3
    BLOG_AUDIO = 5
    MESSAGE = 8
    MESSAGE_AUDIO = 10
    USER_NAME_CARD_BACKGROUND = 11



from dataclasses import dataclass
from dataclasses_json import dataclass_json
from dataclasses_json import LetterCase
from typing import Optional
from datetime import datetime
from .parse import time_field
from .blog import Blog
from .chat import Chat


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class BlockedItemWrapper:
    id: Optional[int] = None
    object_id: Optional[int] = None
    object_type: Optional[int] = None
    chat: Optional[Chat] = None
    blog: Optional[Blog] = None
    created_time: Optional[datetime] = time_field()




from argparse import ArgumentParser
from asyncio import get_event_loop
from asyncio import create_task
from asyncio import gather
from inspect import iscoroutinefunction
from sys import exit as sys_exit
from aiofiles import open as async_open
from os import remove
from . import *

parser = ArgumentParser(description="ProjZ.py library command line interface")
parser.add_argument("action", type=str, help="The action to be performed. "
                                             "You can view the list of possible actions by "
                                             "setting the value \"list-actions\"")
parser.add_argument("--login", type=str, dest="login", help="Email/Phone number of the account")
parser.add_argument("--password", type=str, dest="password", help="Password of the account")
parser.add_argument("--auth", type=str, dest="auth_type", help="Auth method: email or phone")
parser.add_argument("--link", type=str, dest="link", help="Link to the object to get information about")
parser.add_argument("--thread", type=str, dest="thread", help="Link to the chat to send message")
parser.add_argument("--message", type=str, dest="message", help="Text of the message to send")
parser.add_argument("--repeat", type=int, dest="repeat", help="The number of messages sent")
parser.add_argument("--circle", type=str, dest="circle", help="Link to the circle to join or leave")
parser.add_argument(
    "-fLH",
    "--flag-logging-http",
    action="store_true",
    dest="flag_http_logging",
    default=False,
    help="If True, the console will display information about the HTTP requests being sent."
)
parser.add_argument(
    "-fLW",
    "--flag-logging-websocket",
    action="store_true",
    dest="flag_ws_logging",
    default=False,
    help="If True, the console will display information about the websocket messages."
)
parser.add_argument(
    "-fSA",
    "--flag-save-auth",
    action="store_true",
    dest="flag_save_auth",
    default=False,
    help="If True, save the auth data to the cli-auth.json file"
)
cli_args = parser.parse_args()

client = Client(http_logging=cli_args.flag_http_logging, ws_logging=cli_args.flag_ws_logging)


def run_function(func: Callable, *args, **kwargs) -> Any:
    return get_event_loop().run_until_complete(func(*args, **kwargs))\
        if iscoroutinefunction(func) else func(*args, **kwargs)


def print_error(func: Callable):
    def wrapper(*args, **kwargs) -> Any:
        try:
            return run_function(func, *args, **kwargs)
        except Exception as e:
            print(f"Error: {e}")

    return wrapper


def with_auth(get_auth_response: bool):
    def decorator(func: Callable):
        async def wrapper(*args, **kwargs):
            saved_auth_data = await read_auth_file()

            result: AuthResult
            if saved_auth_data is not None:
                try:
                    result = await login(secret_str=saved_auth_data.secret)
                except projz.error.ApiException:
                    remove_auth_file()
                    return await wrapper(*args, **kwargs)
            else:
                if cli_args.login is None or cli_args.password is None or cli_args.auth_type is None:
                    print("Error: A file with the valid auth data is not found, "
                          "please specify login, password and auth type.")
                    return
                result = await login(cli_args.auth_type, cli_args.login, cli_args.password)

            if cli_args.flag_save_auth:
                await write_auth_file(result)

            if get_auth_response:
                args = (result,) + args

            return await func(*args, **kwargs)
        return wrapper

    return decorator


def with_args(*needed_cli_args):
    def decorator(func: Callable):
        def wrapper(*args, **kwargs):
            function_args = []
            for arg in needed_cli_args:
                matching_cli_arg = cli_args.__dict__.get(arg)
                if matching_cli_arg is None:
                    print(f"Error: The required argument \"{arg}\" is not specified")
                    return
                function_args.append(matching_cli_arg)
            args += tuple(function_args)

            return run_function(func, *args, **kwargs)
        return wrapper

    return decorator


async def login(
    auth_type_str: Optional[str] = None,
    login_str: Optional[str] = None,
    password_str: Optional[str] = None,
    secret_str: Optional[str] = None
) -> AuthResult:
    if auth_type_str == "email":
        return await client.login_email(login_str, password_str)
    elif auth_type_str == "phone":
        return await client.login_phone_number(login_str, password_str)
    else:
        return await client.login_secret(secret_str)


async def write_auth_file(auth: AuthResult):
    async with async_open("cli-auth.json", "w") as file:
        await file.write(auth.to_json(indent=4))


async def read_auth_file() -> AuthResult:
    try:
        file = await async_open("cli-auth.json", "r")
        data = AuthResult.from_json(await file.read())
        await file.close()
        return data
    except Exception as e:
        return None


def remove_auth_file():
    remove("cli-auth.json")


async def safe_send_chat_message(chat_id: int, message_content: str) -> bool:
    try:
        await client.send_message(chat_id, content=message_content)
    except:
        return False
    return True


def action_list_actions():
    for i, key in enumerate(action_mapping, start=1):
        print(f"{i}. {key} - {action_mapping[key][1]}")


def action_clear_auth():
    remove_auth_file()
    print("[+] Cleared successfully.")


@print_error
@with_auth(get_auth_response=True)
async def action_login(auth: AuthResult):
    print(f"[+] Session ID: {auth.sid}")
    print(f"[+] Refresh token (secret): {auth.secret}")
    print("[+] Account:")
    print(f"    -- Email: {auth.account.email or '-'}")
    print(f"    -- Phone number: {auth.account.phone_number or '-'}")
    print(f"    -- Registration time: {auth.account.created_time or '-'}")
    print(f"    -- Registered IPv4: {auth.account.registered_ipv4 or '-'}")
    print(f"    -- Last login IPv4: {auth.account.last_login_ipv4 or '-'}")
    print("[+] User Profile: ")
    print(f"    -- User ID: {auth.user_profile.uid}")
    print(f"    -- Nickname: {auth.user_profile.nickname}")
    print(f"    -- Tagline: {auth.user_profile.tagline or '-'}")
    print(f"    -- Bio: {auth.user_profile.bio or '-'}")
    if auth.user_profile.icon is not None:
        print("    -- Icon resources:")
        for resource in auth.user_profile.icon.resource_list:
            print(f"        -- URL: {resource.url}; width: {resource.width}; height: {resource.height}")
    if auth.user_profile.name_card_background is not None:
        print("    -- Name card background resources:")
        for resource in auth.user_profile.name_card_background.resource_list:
            print(f"        -- URL: {resource.url}; width: {resource.width}; height: {resource.height}")


@print_error
@with_args("link")
@with_auth(get_auth_response=False)
async def action_link_info(link: str):
    info = await client.get_link_info(link)
    print(f"[+] Path: {info.path or '-'}")
    print(f"[+] Object ID: {info.object_id or '-'}")
    print(f"[+] Object type: {info.object_type or '-'}")
    print(f"[+] Share link: {info.share_link or '-'}")
    print(f"[+] Parent type: {info.parent_type or '-'}")


@print_error
@with_args("thread", "message", "repeat")
@with_auth(get_auth_response=False)
async def action_send_message(chat_link: str, message_content: str, repeat: int):
    chat_id = (await client.get_link_info(chat_link)).object_id
    results = await gather(*[create_task(safe_send_chat_message(chat_id, message_content)) for _ in range(repeat)])
    sent = list(filter(lambda x: x, results))

    print(f"[{'+' if any(results) else '-'}] Sent: {len(sent)}/{len(results)}")


@print_error
@with_args("circle")
@with_auth(get_auth_response=False)
async def action_join_circle(circle_link: str):
    await client.join_circle(circle_link)
    print(f"[+] Joined successfully.")


@print_error
@with_args("circle")
@with_auth(get_auth_response=False)
async def action_leave_circle(circle_link: str):
    await client.leave_circle(circle_link)
    print(f"[+] Left successfully.")


@print_error
@with_auth(get_auth_response=False)
async def action_listen():
    client.register_chat_message_handler(
        lambda x: print(f"{x.author.nickname}: {x.content}"),
        lambda x: x.content is not None and x.author is not None
    )
    print(f"[+] Waiting for the messages...")

action_mapping = {
    "list-actions": [action_list_actions, "Print a list of all possible actions."],
    "clear-auth": [action_clear_auth, "Remove auth file."],
    "login": [action_login, "Login to the account and print account information."],
    "link-info": [action_link_info, "Get information about the object by the link."],
    "send-message": [action_send_message, "Send message to the chat."],
    "join-circle": [action_join_circle, "Join to the circle by the link."],
    "leave-circle": [action_leave_circle, "Leave from the circle by the link."],
    "listen": [action_listen, "Listen for the chat messages."]
}


def main():
    if cli_args.action not in action_mapping:
        print("Unknown action.")
        return

    fn = action_mapping[cli_args.action][0]

    run_function(fn)
    if cli_args.action == "listen":
        try:
            get_event_loop().run_forever()
        except KeyboardInterrupt:
            print("Interrupted.")
            sys_exit(0)


if __name__ == "__main__":
    main()



from dataclasses import dataclass
from dataclasses_json import dataclass_json
from dataclasses_json import LetterCase
from datetime import datetime
from typing import Optional
from .currency import Currency
from .gift_box import GiftBox
from .user import User
from .parse import extensions_field
from .parse import time_field
from .parse import wallet_amount_field


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class TransferOrder:
    order_id: Optional[int] = None
    order_type: Optional[int] = None
    currency_type: Optional[int] = None
    order_status: Optional[int] = None
    from_uid: Optional[int] = None
    to_uid: Optional[int] = None
    duration: Optional[int] = None
    from_user: Optional[User] = None
    to_user: Optional[User] = None
    currency: Optional[Currency] = None
    gift_box: Optional[GiftBox] = None
    amount: Optional[int] = wallet_amount_field()
    min_claimed_time: Optional[datetime] = time_field()
    created_time: Optional[datetime] = time_field()
    expired_time: Optional[datetime] = time_field()
    extensions: dict = extensions_field()



from enum import Enum


class EContentRegion(Enum):
    ALL = 0
    ENGLISH = 1
    ARABIC = 2
    RUSSIAN = 3
    PORTUGUESE = 4
    SPANISH = 5
    GERMAN = 6
    FRENCH = 7
    BURMESE = 8
    OTHER = 100



from typing import Optional


class ApiException(Exception):

    errors = {}

    def __init__(self, code: int, message: str, *args, **kwargs):
        self.code = code
        self.message = message
        super().__init__(self.message)

    @classmethod
    def create(cls, code: int, extras: Optional[dict[str, str]] = None):
        def decorator(clazz: type):
            def init_wrapper(instance, message: str, response: Optional[dict] = None):
                if response is None: response = dict()
                cls.__init__(instance, code, message, response)
                if extras is not None:
                    for extra in extras: setattr(instance, extras[extra], response.get(extra))
            clazz.__init__ = init_wrapper
            cls.errors[code] = clazz
            return clazz
        return decorator

    @classmethod
    def get(cls, response: dict) -> "ApiException":
        code = response.get("apiCode") or -1
        message = response.get("apiMsg") or ""
        found_error = (cls.errors.get(code) or cls)
        if found_error is ApiException:
            return ApiException(code, message)
        return found_error(message, response)


@ApiException.create(-1)
class BadResponse(ApiException): ...


@ApiException.create(2009)
class InvalidEmail(ApiException): ...


@ApiException.create(1008, extras={"redirectUrl": "url"})
class CaptchaCaught(ApiException): ...


@ApiException.create(2005)
class InvalidVerificationCode(ApiException): ...


@ApiException.create(1000004)
class AccountHasNoWallet(ApiException): ...



from ..api import RequestManager
from ..error import ApiException
from ..enum import EWebSocketEventType
from ..util import SubscriptionHandler
from ..model import ChatMessage
from aiohttp import ClientSession
from aiohttp import ClientWebSocketResponse
from aiohttp import WSMsgType
from asyncio import create_task
from asyncio import sleep
from asyncio import get_running_loop
from typing import Optional
from ujson import loads
from ujson import dumps
from datetime import datetime


class WebsocketListener(SubscriptionHandler):
    def __init__(
        self,
        request_manager: RequestManager,
        logging: bool = False
    ):
        super().__init__()
        self.request_manager = request_manager
        self.provider = request_manager.provider
        self.language = request_manager.language
        self.country_code = request_manager.country_code
        self.time_zone = request_manager.time_zone
        self.device_id = request_manager.device_id
        self.connection: ClientWebSocketResponse = None
        self.task_receiver = None
        self.task_pinger = None
        self.client_session = None
        self.logging = logging
        self.outgoing = {}

    def _log(self, log_type: str, message_type: int, content_length: int):
        log_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        print(f"[WebSocket {log_time}] [{log_type}] [{message_type}] [{content_length} bytes]")

    async def connect(self):
        self.client_session = ClientSession(base_url="wss://ws.projz.com")
        self.connection = await self.client_session.ws_connect(
            "/v1/chat/ws",
            headers=await self.request_manager.build_headers("/v1/chat/ws")
        )
        self.task_receiver = create_task(self.receive())
        self.task_pinger = create_task(self.ping())

    async def disconnect(self):
        self.task_receiver.cancel()
        self.task_pinger.cancel()
        await self.connection.close()
        self.connection = None
        await self.client_session.close()
        self.client_session = None

    async def receive(self):
        while True:
            msg = await self.connection.receive()
            if msg.type != WSMsgType.TEXT: continue
            msg_json = loads(msg.data)
            if self.logging:
                self._log("INCOMING", msg_json["t"], len(msg.data))
            if msg_json["t"] == EWebSocketEventType.MESSAGE.value:
                self.broadcast(ChatMessage.from_dict(msg_json["msg"]))
            elif msg_json["t"] == EWebSocketEventType.ACK.value:
                ack = msg_json["serverAck"]
                if ack["seqId"] not in self.outgoing: continue
                if ack["apiCode"] != 0:
                    self.outgoing[ack["seqId"]].set_exception(ApiException.get(ack))
                    continue
                self.outgoing[ack["seqId"]].set_result(ack)
                del self.outgoing[ack["seqId"]]

    async def ping(self):
        while True:
            await sleep(3)
            await self.send_request(8)

    async def send_request(
        self,
        request_type: int,
        wait_response: bool = False,
        seq_id: Optional[int] = None,
        **kwargs
    ) -> Optional[dict]:
        data = dumps(dict(t=request_type, **kwargs))
        if self.logging:
            self._log("OUTGOING", request_type, len(data))
        await self.connection.send_str(data)
        if wait_response:
            if seq_id is None: raise ValueError("Can't wait for response without seq id parameter")
            future = get_running_loop().create_future()
            self.outgoing[seq_id] = future
            return await future
        return None




from typing import Optional


class PaginatedList(list):
    def __init__(self, original_list: list, pagination: Optional[dict] = None, next_page_token: Optional[str] = None):
        super().__init__(original_list)
        self.next_page_token = next_page_token or (None if pagination is None else pagination.get("nextPageToken"))



from setuptools import setup
from setuptools import find_packages
from os import environ
from platform import system
from subprocess import run


def get_readme() -> str:
    with open("README.md", "r") as file:
        return file.read()


def get_requirements() -> list[str]:
    return [
        "aiohttp",
        "aiofiles",
        "ujson",
        "dataclasses",
        "dataclasses-json",
        "python-dateutil",
        "dill",
        "python-magic-bin" if system().lower() == "windows" else "python-magic"
    ]


def main():
    if "PREFIX" in environ and "/com.termux" in environ["PREFIX"]:
        print("Termux detected, checking Sox package...")
        if "sox" not in run(["pkg", "list-installed", "sox"], capture_output=True).stdout.decode("utf-8"):
            print("Sox not found, installing...")
            if run(["pkg", "install", "sox", "-y"], capture_output=True).returncode == 0:
                print("Sox has been successfully installed.")
            else:
                print("Failed to install Sox. The stability of the library is not guaranteed.")
        else:
            print("Sox found, no installation required.")

    setup(
        name="cloverspace",
        version="3.0.0",
        author="LixxRarin",
        description="An asynchronous library for creating scripts and chatbots in Clover.Space;",
        url="https://github.com/LixxRarin/Clover.Space/",
        packages=find_packages(),
        author_email="lixxrarin@gmail.com",
        install_requires=get_requirements(),
        long_description=get_readme(),
        long_description_content_type="text/markdown",
        python_requires=">=3.7",
        keywords=[
            "project-z", "projectz", "projz",
            "bots", "api", "supersymlab",
            "projz-bot", "projz-bots", "projz-api",
            "cloverspace", "clover-space"
        ]
    )


if __name__ == "__main__":
    main()



from dataclasses import dataclass
from dataclasses import field
from dataclasses_json import dataclass_json
from dataclasses_json import LetterCase
from typing import Optional
from .media import Media
from .parse import wallet_amount_field


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Currency:
    currency_type: Optional[int] = None
    name: Optional[str] = None
    symbol: Optional[str] = None
    icon: Optional[Media] = None
    amount: Optional[int] = wallet_amount_field()
    decimals: Optional[int] = field(default_factory=lambda: 18)



from dataclasses import dataclass
from dataclasses_json import dataclass_json
from dataclasses_json import LetterCase
from typing import Optional
from .media import Media


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Dice:
    dice_id: Optional[int] = None
    name: Optional[str] = None
    icon: Optional[Media] = None
    animation: Optional[Media] = None
    sides: Optional[list[Media]] = None



from dataclasses import dataclass
from dataclasses_json import dataclass_json
from dataclasses_json import LetterCase
from dataclasses_json import config
from dataclasses import field
from typing import Optional
from .currency import Currency


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class UserTask:
    type: Optional[int] = None
    currency: Optional[Currency] = None
    gift_box_id: Optional[int] = None
    gift_box_status: Optional[int] = field(default_factory=int)
    completed: Optional[bool] = field(metadata=config(field_name="status", decoder=lambda x: x != 1), default_factory=bool)



from dataclasses_json import dataclass_json
from dataclasses_json import LetterCase
from dataclasses import dataclass
from dataclasses import field
from typing import Optional


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Media:
    @dataclass_json(letter_case=LetterCase.CAMEL)
    @dataclass
    class Resource:
        width: int
        height: int
        url: str
        thumbnail: bool = False
        duration: int = 0
    base_url: Optional[str] = None
    media_id: Optional[int] = None
    resource_list: list[Resource] = field(default_factory=list)



from .headers import IHeadersProvider
from .util import CopyToBufferWriter
from ..error import ApiException
from ..error import BadResponse
from io import BytesIO
from aiohttp import ClientSession
from aiohttp import MultipartWriter
from typing import Optional
from typing import Union
from uuid import uuid4
from ujson import dumps
from ujson import loads
from ujson import JSONDecodeError
from dataclasses_json import DataClassJsonMixin
from urllib.parse import urlencode
from datetime import datetime


class RequestManager:
    def __init__(
        self,
        provider: IHeadersProvider,
        language: str = "en-US",
        country_code: str = "us",
        time_zone: int = 180,
        logging: bool = False
    ):
        self.provider = provider
        self.language = language
        self.country_code = country_code
        self.time_zone = time_zone
        self.device_id = None
        self.logging = logging

    async def build_headers(self, endpoint: str, body: Optional[bytes] = None, extra: Optional[dict] = None) -> dict:
        if self.device_id is None:
            self.device_id = await self.provider.generate_device_id(str(uuid4()))
        headers = self.provider.get_persistent_headers()
        headers.update(self.provider.get_request_info_headers(
            self.device_id,
            str(uuid4()),
            self.language,
            self.country_code,
            self.time_zone
        ))
        headers.update(extra or {})
        headers["HJTRFS"] = await self.provider.generate_request_signature(endpoint, headers, body or bytes())
        return headers

    async def request(
        self,
        method: str,
        endpoint: str,
        params: Optional[dict] = None,
        body: Optional[bytes] = None,
        content_type: Optional[str] = None,
        web: bool = True
    ) -> dict:
        if not endpoint.startswith("/"): endpoint = f"/{endpoint}"
        if params: endpoint += f"?{urlencode(params)}"
        if web: endpoint = f"/api/f{endpoint}"
        if self.logging:
            request_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            print(
                f"[HTTP {request_time}] " +
                (f"[{method} {endpoint}]" if body is None else f"[{method} {endpoint}] [{len(body)} bytes]")
            )
        async with ClientSession(base_url="https://api.projz.com" if not web else "https://www.projz.com") as session:
            response = await session.request(
                method,
                endpoint,
                headers=await self.build_headers(
                    endpoint,
                    body,
                    {"Content-Type": content_type} if content_type is not None else None
                ) if not web else dict(),
                data=body
            )
            try: response_json = loads(await response.text())
            except (JSONDecodeError, UnicodeDecodeError): raise BadResponse("Can't read response from Project Z API")
            if "apiCode" in response_json: raise ApiException.get(response_json)
            return response_json

    async def get(self, endpoint: str, params: Optional[dict] = None, web: bool = False) -> dict:
        return await self.request("GET", endpoint, params=params or dict(), web=web)

    async def delete(self, endpoint: str, params: Optional[dict] = None, web: bool = False) -> dict:
        return await self.request("DELETE", endpoint, params=params or dict(), web=web)

    async def post(
        self,
        endpoint: str,
        body: Union[bytes, str, dict, DataClassJsonMixin, MultipartWriter],
        content_type: Optional[str],
        web: bool = False
    ) -> dict:
        content = BytesIO()
        if isinstance(body, bytes): content.write(body)
        elif isinstance(body, str): content.write(body.encode("utf-8"))
        elif isinstance(body, dict): content.write(dumps(body).encode("utf-8"))
        elif isinstance(body, DataClassJsonMixin): content.write(body.to_json().encode("utf-8"))
        elif isinstance(body, MultipartWriter): await body.write(CopyToBufferWriter(content))
        else: raise ValueError(f"Invalid request body type: \"{body.__class__.__name__}\"")
        return await self.request("POST", endpoint, body=content.getvalue(), content_type=content_type, web=web)

    async def post_json(self, endpoint: str, body: Union[dict, DataClassJsonMixin], web: bool = False) -> dict:
        return await self.post(endpoint, body=body, content_type="application/json; charset=UTF-8", web=web)

    async def post_empty(self, endpoint: str, web: bool = False):
        return await self.post(endpoint, body="", content_type=None, web=web)



from .base_headers_provider import IHeadersProvider
from hashlib import sha1
from hashlib import sha256
from hmac import HMAC
from base64 import b64encode
from time import time


class HeadersProvider(IHeadersProvider):
    def __init__(self):
        self.sid = ""

    def get_persistent_headers(self) -> dict:
        return {
            "appType": "MainApp", "appVersion": "2.27.1",
            "osType": "2", "deviceType": "1", "flavor": "google",
            "User-Agent": "com.projz.z.android/2.27.1-25104 (Linux; U; Android 7.1.2; ASUS_Z01QD; Build/Asus-user 7.1.2 2017)"
        }

    def get_request_info_headers(self, device_id: str, nonce: str, language: str, country_code: str, time_zone: int) -> dict:
        headers = {
            "rawDeviceId": device_id,
            "nonce": nonce,
            "Accept-Language": language,
            "countryCode": country_code.upper(),
            "carrierCountryCodes": country_code,
            "timeZone": str(time_zone),
            "reqTime": str(int(time() * 1000))
        }
        if self.sid is not None: headers["sId"] = self.sid
        return headers

    def get_signable_header_keys(self) -> list[str]: return [
        "rawDeviceId", "rawDeviceIdTwo", "rawDeviceIdThree",
        "appType", "appVersion", "osType",
        "deviceType", "sId", "countryCode",
        "reqTime", "User-Agent", "contentRegion",
        "nonce", "carrierCountryCodes"
    ]

    def get_sid(self) -> str: return self.sid

    def set_sid(self, sid: str) -> None: self.sid = sid

    def remove_sid(self) -> None: self.sid = ""

    async def generate_request_signature(self, path: str, headers: dict, body: bytes) -> str:
        mac = HMAC(key=bytes.fromhex("ce070279278de1b6390b76942c13a0b0aa0fda6aedd6f2d655eda7cf6543b35f" + ("6a" * 32)),
                   msg=path.encode("utf-8"),
                   digestmod=sha256)
        for header in [headers[signable] for signable in self.get_signable_header_keys() if signable in headers]:
            mac.update(header.encode("utf-8"))
        if body: mac.update(body)
        return b64encode(bytes.fromhex("04") + mac.digest()).decode("utf-8")

    async def generate_device_id(self, installation_id: str) -> str:
        prefix = bytes.fromhex("04") + sha1(installation_id.encode("utf-8")).digest()
        return (
            prefix + sha1(
                prefix + sha1(bytes.fromhex("997ec928a85f539e3fa124761e7572ef852e")).digest()
            ).digest()
        ).hex()

    async def generate_device_id_three(
        self,
        organization: str,
        platform: str,
        version: str,
        model: str,
        app_id: str
    ) -> str:
        raise NotImplementedError



from .gender import EGender
from .upload_target import EUploadTarget
from .ws_event_type import EWebSocketEventType
from .chat_thread_type import EChatThreadType
from .chat_message_type import EChatMessageType
from .content_region import EContentRegion
from .object_type import EObjectType
from .circle_filter_type import ECircleFilterType
from .auth_type import EAuthType
from .auth_purpose import EAuthPurpose
from .user_task_type import EUserTaskType
from .currency_type import ECurrencyType
from .party_query_type import EPartyQueryType
from .chat_query_type import EChatQueryType
from .circle_members_query_type import ECircleMembersQueryType



from dataclasses import dataclass
from dataclasses import field
from dataclasses_json import dataclass_json
from dataclasses_json import LetterCase
from .user import User
from .chat import Chat
from typing import Optional


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Party:
    admin_user: Optional[User] = None
    chat_thread_list: list[Chat] = field(default_factory=list)



from dataclasses import dataclass
from dataclasses import field
from dataclasses_json import dataclass_json
from dataclasses_json import LetterCase
from dataclasses_json import config
from datetime import datetime
from typing import Optional
from .user import User
from .parse import extensions_field
from .parse import time_field
from .parse import wallet_amount_field


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class GiftBox:
    box_id: Optional[int] = None
    uid: Optional[int] = None
    to_object_id: Optional[int] = None
    to_object_type: Optional[int] = None
    currency_type: Optional[int] = None
    max_claimed_count: Optional[int] = None
    distribute_mode: Optional[int] = None
    accepted: Optional[bool] = field(metadata=config(field_name="giftBoxStatus", decoder=lambda x: x != 1), default_factory=bool)
    title: Optional[str] = None
    user: Optional[User] = None
    claimed_amount: Optional[int] = wallet_amount_field()
    expired_time: Optional[datetime] = time_field()
    created_time: Optional[datetime] = time_field()
    amount: Optional[int] = wallet_amount_field()
    extensions: dict = extensions_field()

